/*
 * reconfigureClient.h
 *
 *  Created on: 23/06/2016
 *      Author: rescue
 */

#ifndef CROSBOT_RECONFIGURE_CLIENT_H_
#define CROSBOT_RECONFIGURE_CLIENT_H_

#include <crosbot/handle.hpp>
#include <crosbot/thread.hpp>

#include <dynamic_reconfigure/server.h>
#include <dynamic_reconfigure/Config.h>

#include <boost/config.hpp>

namespace crosbot {

/**
 * C++ Client for allowing reconfiguration of a given node's parameters
 * through ROS dynamic_reconfigure.
 * The header file containing the class for the node's config must be placed
 * above this header file to ensure compilation.
 *
 * @tparam TConfig Dynamic Reconfiguration class that the client updates
 */
template <class TConfig>
class ReconfigureClient : public HandledObject {

public:
    /**
     * Callback for parameter updates
     */
    void callback_paramaterUpdates(dynamic_reconfigure::Config config) {
        Lock configLock(configMutex);
        latestConfig = config;
        recievedConfig = true;
    };

    /**
     * Create configuration client and subscribe to dynamic_reconfigure publishers and services.
     * Assumes topic names are /node/set_parameters
     * and /node/parameter_updates
     *
     * @param node Name of node to reconfigure
     */
    ReconfigureClient(std::string node) :
        node(node),
        recievedConfig(false)
    {
        std::string serviceName = "/" + node + "/set_parameters";
        std::string updatesName = "/" + node + "/parameter_updates";

        // Connect to existing service
        ros::NodeHandle nh("~");
        service = nh.serviceClient<dynamic_reconfigure::ReconfigureRequest>(serviceName);
        subscriber = nh.subscribe<dynamic_reconfigure::Config>(updatesName, 1, &ReconfigureClient<TConfig>::callback_paramaterUpdates, this);
    };

    /**
     * Destroy client, unsubscribe from publishers and services.
     */
    virtual ~ReconfigureClient() {
        service.shutdown();
        subscriber.shutdown();
        recievedConfig = false;
    }

    /**
     * Check to see if a set of configuration parameters have been received.
     *
     * @return true if the latest configuration parameters have been received.
     */
    bool hasLatestConfig() {
        return recievedConfig;
    }

    /**
     * Return (in TConfig format) the latest configuration as found on the parameter server.
     * Should only be called if hasLatestConfig returns true.
     *
     * @param tconfig Reference to configuration data structure to load with config
     */
    void getLatestConfig(TConfig& tconfig) {
        Lock configLock(configMutex);
        tconfig.__fromMessage__(latestConfig);
    }

    /**
     * Update new set of parameters
     */
    void updateParameters(TConfig& config) {
        ROS_INFO("update reconfigure parameters");
        Lock configLock(configMutex);

        // Service request/response
        dynamic_reconfigure::ReconfigureRequest request;
        dynamic_reconfigure::ReconfigureResponse response;

        // Call service
        if (service.waitForExistence(ros::Duration(0.5))) {
            config.__toMessage__(request.config);
            service.call(request, response);
        } else {
            ROS_WARN("Service '%s' not available", service.getService().c_str());
        }

        // Update latest config response
        latestConfig = response.config;
    }
private:
    // Fields
    std::string node;
    bool recievedConfig;
    dynamic_reconfigure::Config latestConfig;

    // ROS publishers and services
    ros::ServiceClient service;
    ros::Subscriber subscriber;

    // Multi-threading
    Mutex configMutex;
};

#ifndef BOOST_NO_CXX11_TEMPLATE_ALIASES

template <class TConfig>
using ReconfigureClientPtr = Handle<ReconfigureClient<TConfig> >;

#endif

} // namespace crosbot

#endif /* CROSBOT_RECONFIGURE_CLIENT_H_ */
