/*
 * explorer.hpp
 *
 *  Created on: 22/02/2013
 *      Author: mmcgill
 */

#ifndef CROSBOT_EXPLORER_HPP_
#define CROSBOT_EXPLORER_HPP_

#include <crosbot/thread.hpp>
#include <crosbot_map/voronoi.hpp>

#include <crosbot_explore/actionServer/actionParameters.hpp>
#include <crosbot_explore/driveParameters.hpp>
#include <crosbot_explore/hysteresisData.hpp>
#include <crosbot_explore/hysteresisParameters.hpp>
#include <crosbot_explore/searchParameters.hpp>

namespace crosbot {

class Explorer {
protected:
	void planThread();
	void driveThread();

	class PlanThread : public Thread {
	public:
		Explorer& explorer;
		PlanThread(Explorer& explorer) : explorer(explorer) {}

		void run() {
			explorer.planThread();
		}
	};

	class DriveThread : public Thread {
	public:
		Explorer& explorer;
		DriveThread(Explorer& explorer) : explorer(explorer) {}

		void run() {
			explorer.driveThread();
		}
	};

	// Processing threads
	PlanThread planThread_;
	DriveThread driveThread_;

	// Concurrency handling
	Mutex planningThreadMutex;
	Mutex drivingThreadMutex;

	// Configuration
	bool debugMsgs;
	DriveParameters driveParams;
	HysteresisParameters hysteresisParams;
	SearchParameters searchParams;
	bool searchWhilePaused;
	VoronoiGrid::Constraints voronoiConstraints;

	// Processing data
	bool operating;
	bool paused;
	bool stopSent;
	HysteresisData hysteresis;

	// Searching for drive targets
	virtual Pose findDriveTarget(const VoronoiGrid& voronoi, const Pose& robot);
	virtual Pose findWallFollowTarget(const VoronoiGrid& voronoi, const Pose& robot);
	virtual Pose findWaypointTarget(const VoronoiGrid& voronoi, const Pose& robot);
	virtual ImagePtr getPlanImage(const VoronoiGrid& voronoi, const Pose& robot, const Pose& target);

	// Hysteresis of finding drive target
	virtual Pose findHysteresisTarget(const VoronoiGrid& voronoi, const Pose& robot, const Pose& target);
	void hysteresisPreprocess(const Time& timestamp, const Pose& robot);
	int hysteresisAssignCluster(const Time& timestamp, const Pose& target, const Pose& robot);
	int hysteresisIndex(int assignedCluster, const Time& timestamp);

	bool cellIsDirectlyTraversible(const VoronoiGrid& voronoi, const Index2D& cell, const Index2D& pov) const;
	double cellTraversibleRisk(const VoronoiGrid& voronoi, const Index2D& cell, const Index2D& pov) const;

	// functions for the wall follow implementation
	virtual Point findWall(const VoronoiGrid& voronoi, const Pose& robot, double startAngle);
	virtual Index2D findFirstSkeletonCell(const VoronoiGrid& voronoi, const Pose& robot, double startAngle);
	virtual Index2D findNextSkeletonCell(const VoronoiGrid& voronoi, const Index2D currentCell, const Index2D previous);
	virtual Index2D findSaferSkeletonCell(const VoronoiGrid& voronoi, const Index2D currentCell, const Index2D robot);
//  virtual Index2D findSaferCell(const VoronoiGrid& voronoi, const Index2D currentCell, const Index2D robot);
//	virtual void findSaferTarget(double& voronoiAngle, double& voronoiDistance, const Pose& robot);

	// functions for the waypoint follow implementation
	virtual double traversibleDistance(const VoronoiGrid& voronoi, const Point position, const double angle, const double max = INFINITY);

	// Calculations
	virtual double poseYawDifference(const Pose& from, const Pose& to);

	// Driving execution
	virtual Pose getLatestPose()=0;
	virtual bool stopMotors()=0;
	virtual void driveTo(const Pose& relativePosition)=0;

public:
	Explorer();
	virtual ~Explorer();

	virtual void shutdown();
	virtual void pause();
	virtual void resume();
	virtual void statusChanged(const std::string& status) {}

	virtual VoronoiGridPtr getLatestVoronoi()=0;

protected:

	// StatusChanged strings
	std::string STATUS_ARRIVED;
	std::string STATUS_BLOCKED;
	std::string STATUS_RUNNING;
	std::string STATUS_PAUSED;
};

} // namespace crosbot

#endif /* CROSBOT_EXPLORER_HPP_ */
