/*
 * controlWrapper.cpp
 *
 *  Created on: 04/07/2016
 *      Author: rescue
 */

#include <ros/ros.h>

#include <crosbot/geometry/poses.hpp>
#include <crosbot_explore/actionServer/actionParameters.hpp>
#include <crosbot_explore/actionServer/basicClient.hpp>
#include <crosbot_explore/actionServer/basicClientCommands.hpp>
#include <crosbot_explore/nodes/astarExplorerNode.hpp>

#include <std_msgs/String.h>

#define CONTROL_COMMAND_RESET_MAP       "reset_map"
#define DEFAULT_SERVER_WAIT             0.5
#define LOG_START                       "BasicExplorerClient ::"


#define STATUS_ERROR(STPUB, FMT, ...)                                                       \
        STPUB->sendStatus(crosbot_msgs::ControlStatus::LEVEL_ERROR, FMT, __VA_ARGS__);      \
        ROS_ERROR(FMT, __VA_ARGS__);

#define STATUS_INFO(STPUB, FMT, ...)                                                        \
        STPUB->sendStatus(crosbot_msgs::ControlStatus::LEVEL_INFO, FMT, __VA_ARGS__);       \
        ROS_INFO(FMT, __VA_ARGS__);

#define STATUS_WARN(STPUB, FMT, ...)                                                        \
        STPUB->sendStatus(crosbot_msgs::ControlStatus::LEVEL_WARNING, FMT, __VA_ARGS__);    \
        ROS_WARN(FMT, __VA_ARGS__);


namespace crosbot_explore {

class _CommandCallback : public crosbot::CrosbotCommandCallback {
private:
    BasicExplorerClientPtr wrapper;

public:
    _CommandCallback(BasicExplorerClientPtr wrapper) : wrapper(wrapper) {};
    virtual ~_CommandCallback() {};
    virtual void callback_receivedCommand(const crosbot_msgs::ControlCommandPtr command) {
        if (wrapper != NULL) wrapper->callback_receivedCommand(command);
    };
};

BasicExplorerClient::BasicExplorerClient() :
    actionClient(NULL)
{}

BasicExplorerClient::~BasicExplorerClient() {
}

void BasicExplorerClient::configure() {
    // Use "local" namespace for config
    ros::NodeHandle nh("~");

    aStarSrvName = nh.param<std::string>("astar_path_topic", "astar_path");
    exploreAServerName = nh.param<std::string>("explore_action_server_topic", "/crosbot/explorer");
    navGoalSub_name = nh.param<std::string>("navGoal_sub", "/move_base_simple/goal");
    robot_frame = nh.param<std::string>("robotFrame", "base_link");
    postrackResetPub_name = nh.param<std::string>("postrackReset_pub", "/postrack/resetMap");
    slamResetPub_name = nh.param<std::string>("slamReset_pub", "/graphslam/resetMap");
    world_frame = nh.param<std::string>("worldFrame", "icp_test");
}

void BasicExplorerClient::startup() {
    // Use "local" namespace for config
    ros::NodeHandle nh("~");

    crosbotCommand = new crosbot::CrosbotCommand(DEFAULT_CONTROL_NAMESPACE, true,
            new _CommandCallback(this));
    crosbotStatus = new crosbot::CrosbotStatus(DEFAULT_CONTROL_NAMESPACE);
    aStarSrv = nh.serviceClient<crosbot_explore::GetPath>(aStarSrvName);
    pubPostrackReset = nh.advertise<std_msgs::String>(postrackResetPub_name, 1);
    pubGraphSlamReset = nh.advertise<std_msgs::String>(slamResetPub_name, 1);
    subNavGoal = nh.subscribe(navGoalSub_name, 1, &BasicExplorerClient::callback_navGoal, this);

    actionClient = new ExploreAClient(exploreAServerName, true);
}

void BasicExplorerClient::shutdown() {
    crosbotCommand = NULL;
    //exploreModeSrv.shutdown();

    actionClient->cancelAllGoals();
    delete actionClient;
}

void BasicExplorerClient::callback_navGoal(const geometry_msgs::PoseStampedConstPtr& targetPose) {
    geometry_msgs::PoseStamped pose = *targetPose;
    setAStarPose(pose, true);
}

void BasicExplorerClient::callback_receivedCommand(const crosbot_msgs::ControlCommandPtr command) {
    if (command->command == crosbot_msgs::ControlCommand::CMD_EM_STOP) {
        setExploreMode(crosbot_explore::ExploreMode::MODE_STOP);
    } else if (command->command == crosbot_msgs::ControlCommand::CMD_RESET) {
        command_reset(command);
    } else if (command->command == crosbot_msgs::ControlCommand::CMD_RESUME) {
        setExploreMode(crosbot_explore::ExploreMode::MODE_RESUME);
    } else if (command->command == crosbot_msgs::ControlCommand::CMD_START) {
        setExploreMode(crosbot_explore::ExploreMode::MODE_RESUME);
    } else if (command->command == crosbot_msgs::ControlCommand::CMD_STOP) {
        setExploreMode(crosbot_explore::ExploreMode::MODE_STOP);
    } else if (command->command == COMMAND_LEFTWALL_FOLLOW) {
        setExploreMode(crosbot_explore::ExploreMode::MODE_LEFT_WALL);
    } else if (command->command == COMMAND_RIGHTWALL_FOLLOW) {
        setExploreMode(crosbot_explore::ExploreMode::MODE_RIGHT_WALL);
    } else if (command->command == COMMAND_GO_TO_ORIGIN) {
        setAStarPose(crosbot::Pose3D(), false);
    } else if (command->command == COMMAND_GO_TO_POINT) {
        crosbot::Pose3D toPose = getCommandPoint(command);
        if (toPose.isFinite()) {
            setAStarPose(toPose, false);
        } else {
            STATUS_ERROR(crosbotStatus, "%s Go to Point failed - invalid point specified", LOG_START);
        }
    } else if (command->command == COMMAND_GO_TO_POSE) {
        crosbot::Pose3D toPose = getCommandPose(command);
        if (toPose.isFinite()) {
            setAStarPose(toPose, true);
        } else {
            STATUS_ERROR(crosbotStatus, "%s Go to Pose failed - invalid point specified", LOG_START);
        }
    }
}

void BasicExplorerClient::actionserver_goalActive() {
}

void BasicExplorerClient::actionserver_goalReached(const actionlib::SimpleClientGoalState& state,
            const crosbot_explore::ExploreResultConstPtr& result) {
    ROS_INFO("%s Finished in state [%s]", LOG_START, state.toString().c_str());
    ROS_INFO("%s Explorer: Reached goal with success: %d", LOG_START, result->success);
}

void BasicExplorerClient::actionserver_feedback(const crosbot_explore::ExploreFeedbackConstPtr& feedback) {
    std::string searchStrategy = "";
    if (feedback->searchStrategy == crosbot_explore::ExploreMode::MODE_LEFT_WALL) {
        searchStrategy = "left wall";
    } else if (feedback->searchStrategy == crosbot_explore::ExploreMode::MODE_RIGHT_WALL) {
        searchStrategy = "right wall";
    } else if (feedback->searchStrategy == crosbot_explore::ExploreMode::MODE_WAYPOINT) {
        searchStrategy = "waypoint";
    }

    STATUS_INFO(crosbotStatus, "Explorer: %s (%s)", feedback->status.c_str(), searchStrategy.c_str());
}

crosbot::Pose3D BasicExplorerClient::getCommandPoint(const crosbot_msgs::ControlCommandPtr command) {
    crosbot::Pose3D toPoint(INFINITY, INFINITY, INFINITY);

    if (command->args_doubles.size() >= 3) {
        toPoint.position.x = command->args_doubles[0];
        toPoint.position.y = command->args_doubles[1];
        toPoint.position.z = command->args_doubles[2];
    }

    return toPoint;
}

crosbot::Pose3D BasicExplorerClient::getCommandPose(const crosbot_msgs::ControlCommandPtr command) {
    crosbot::Pose3D toPose = getCommandPoint(command);

    if (command->args_doubles.size() >= 7) {
        toPose.orientation.x = command->args_doubles[3];
        toPose.orientation.y = command->args_doubles[4];
        toPose.orientation.z = command->args_doubles[5];
        toPose.orientation.w = command->args_doubles[6];
    }

    return toPose;
}

void BasicExplorerClient::setExploreMode(int mode) {
    // Ensure action server has started
    bool hasServer = actionClient->waitForServer(ros::Duration(DEFAULT_SERVER_WAIT));
    if (hasServer) {
        STATUS_INFO(crosbotStatus, "%s sending explore mode: %d", LOG_START, mode);

        crosbot_explore::ExploreGoal goal;
        goal.mode = mode;
        actionClient->sendGoal(goal,
                boost::bind(&BasicExplorerClient::actionserver_goalReached, this, _1, _2),
                boost::bind(&BasicExplorerClient::actionserver_goalActive, this),
                boost::bind(&BasicExplorerClient::actionserver_feedback, this, _1));
    } else {
        STATUS_ERROR(crosbotStatus, "%s Explorer Action Server not available", LOG_START);
    }
}

// TODO: use pose
void BasicExplorerClient::setAStarPose(crosbot::Pose3D targetPose, bool targetOrientation) {
    ros::Time tNow = ros::Time::now();

    // Get A* path - points must be in the frame of reference of the local or world map
    geometry_msgs::PointStamped fromPoint;
    geometry_msgs::PointStamped toPoint;
    fromPoint.point = crosbot::Point3D().toROS();
    toPoint.point = crosbot::Point3D().toROS();

    // Path
    crosbot::Pose3D toPose(crosbot::Point3D(toPoint.point), crosbot::Quaternion());
    geometry_msgs::Pose geomToPose = toPose.toROS();
    geometry_msgs::PoseStamped stamptedToPose;
    stamptedToPose.pose = geomToPose;
    stamptedToPose.header.stamp = tNow;
    stamptedToPose.header.frame_id = world_frame;
    setAStarPose(stamptedToPose, targetOrientation);
}

void BasicExplorerClient::setAStarPose(const geometry_msgs::PoseStamped& targetPose, bool targetOrientation) {
    // Send new action
    bool hasServer = actionClient->waitForServer(ros::Duration(DEFAULT_SERVER_WAIT));
    if (hasServer) {
        int mode = crosbot_explore::MoveMode::MODE_ASTAR;
        STATUS_INFO(crosbotStatus, "%s A* to pose: %s", LOG_START, crosbot::Pose3D(targetPose).position.toString().c_str());

        // Set goal
        crosbot_explore::ExploreGoal goal;
        goal.mode = mode;
        goal.targetOrientation = targetOrientation;

        // Path
        goal.path.header.frame_id = targetPose.header.frame_id;
        goal.path.header.stamp = targetPose.header.stamp;
        goal.path.poses.push_back(targetPose);

        // Send action
        actionClient->sendGoal(goal,
                boost::bind(&BasicExplorerClient::actionserver_goalReached, this, _1, _2),
                boost::bind(&BasicExplorerClient::actionserver_goalActive, this),
                boost::bind(&BasicExplorerClient::actionserver_feedback, this, _1));
    } else {
        STATUS_ERROR(crosbotStatus, "%s Explorer Action Server not available", LOG_START);
    }
}

void BasicExplorerClient::command_reset(const crosbot_msgs::ControlCommandPtr command) {
    std_msgs::String msg;
    msg.data = CONTROL_COMMAND_RESET_MAP;
    pubPostrackReset.publish(msg);
    pubGraphSlamReset.publish(msg);
}

} // namespace crosbot_explore


