/*
 * astar.cpp
 *
 *  Created on: 22/02/2013
 *      Author: mmcgill
 */

#include <ros/ros.h>

#include <crosbot_explore/explorerROS.hpp>

#include <crosbot/config.hpp>
#include <geometry_msgs/Twist.h>
#include <std_msgs/String.h>


#define DEFAULT_BASEFRAME			    "/base_link"
#define DEFAULT_DRIVE_OPERATING_RATE    100
#define LOG_START                       "crosbot_explore ::"

namespace crosbot {

ExplorerROSNode::ExplorerROSNode() :
    baseFrame(DEFAULT_BASEFRAME),
    actionServer(NULL),
    gotReconfig(false),
    reconfig_server(NULL),
    currentGoalId(-1)
{
    STATUS_ARRIVED = crosbot_explore::ExploreStatus::STATUS_ARRIVED;
    STATUS_BLOCKED = crosbot_explore::ExploreStatus::STATUS_BLOCKED;
    STATUS_RUNNING = crosbot_explore::ExploreStatus::STATUS_RUNNNING;
    STATUS_PAUSED = crosbot_explore::ExploreStatus::STATUS_PAUSED;

    planThread_.start();
    driveThread_.start();
}

ExplorerROSNode::~ExplorerROSNode() {
};

void ExplorerROSNode::callbackOccGrid(const nav_msgs::OccupancyGridConstPtr& latestMap) {
    Lock lock(rosLock, true);
    this->latestMap = latestMap;
}

void ExplorerROSNode::callbackHistory(const nav_msgs::PathConstPtr& latestHistory) {
    Lock lock(rosLock, true);
    this->latestHistory = latestHistory;
}

void ExplorerROSNode::callback_reconfigure(crosbot_explore::CrosbotExploreReconfigConfig &config, uint32_t level) {
    // Lock loading of parameters
    Lock planningLock(planningThreadMutex);

    localVoronoiConstraints.restricted = config.voronoi_restricted;
    localVoronoiConstraints.partial = config.voronoi_partial;
    localVoronoiConstraints.expand = config.voronoi_expand;
    localVoronoiConstraints.orphanThreshold = config.voronoi_orphanThreshold;
    ROS_INFO("%s new dynamic config: \t restrict: %.2lf, partial %.2lf, expand: %.2lf", LOG_START, localVoronoiConstraints.restricted, localVoronoiConstraints.partial, localVoronoiConstraints.expand);

    searchParams.minTravelDistance = config.min_travel_distance;
    searchParams.obstructionAngle = DEG2RAD(config.obstruction_angle);
    searchParams.proximityAngle = DEG2RAD(config.proximity_angle);
    searchParams.proximityDistance = config.proximity_distance;
    searchParams.skeletonSearchDistance = config.skeleton_search_distance;
    searchParams.wallSearchDistance = config.wall_search_distance;
    searchParams.wallSearchAngle = config.wall_search_angle;

    // Change grid sub
    std::string newGridSub_name = config.map_sub;
    if (newGridSub_name != "" && newGridSub_name != gridSub_name) {
        gridSub_name = newGridSub_name;

        ros::NodeHandle nh;
        subscribeGrid(nh);
    }

    driveParams.maxTurn = config.drive_max_turn;
    driveParams.maxVel = config.drive_max_vel;

    currentGoalId = -1;

    gotReconfig = true;
}

void ExplorerROSNode::callback_explore(const crosbot_explore::ExploreGoalConstPtr& goal) {
    ROS_INFO("%s got exploration action server call. Goal: %d", LOG_START, goal->mode);
    currentGoalId = goal->id;

    // Result of the goal action
    crosbot_explore::ExploreResult result;
    result.id = currentGoalId;

    // Configure and start explorer based on mode
    bool enableDrive = false;

    // Configure base on the goal mode
    handleExploreMode(goal, result, enableDrive);

    // Configure drive
    bool driveOperating = false;
    if (enableDrive) {
        ROS_INFO("%s enabling drive", LOG_START);
        driveOperating = true;
        resume();
    }

    ros::Rate rate(DEFAULT_DRIVE_OPERATING_RATE);

    // Operation loop
    while (driveOperating) {
        //ROS_INFO("%s \t\t checking action server", LOG_START);
        if (!ros::ok() || !actionServer->isActive()|| !operating) {
            ROS_ERROR("%s ROS killed - terminating", LOG_START);
            driveOperating = false;

            result.success = false;
            actionServer->setAborted(result);
        } else if (actionServer->isPreemptRequested()) {
            ROS_INFO("%s Goal Preempted/Cancelled", LOG_START);
            driveOperating = false;

            result.success = false;
            actionServer->setPreempted(result);
        } else if (searchParams.atDestination) {
            ROS_INFO("%s Arrived at Destination", LOG_START);
            driveOperating = false;

            result.success = true;
            actionServer->setSucceeded(result);
        } else if (!driveParams.getDriveTarget().isFinite() && stopSent) {
            ROS_INFO("%s Navigation failed - infinite goal (%d) - motors stopped (%d)", LOG_START, !driveParams.getDriveTarget().isFinite(), stopSent);
            driveOperating = false;

            result.success = false;
            actionServer->setAborted(result);
        } else {
            // action should still execute
            rate.sleep();
        }
    }

    // Stop robot is necessary
    if (enableDrive || goal->mode == crosbot_explore::ExploreMode::MODE_STOP) {
        pause();
    }

    // Give response
    if (!enableDrive) {
        // If drive didn't happen, assume goal succeeded and respond
        actionServer->setSucceeded(result);
    }
}

void ExplorerROSNode::handleExploreMode(const crosbot_explore::ExploreGoalConstPtr& goal,
        crosbot_explore::ExploreResult& result, bool& enableDrive) {
    // Locking concurrency controls is done based on the selected mode

    // Load local voronoi constraints
    if (!gotReconfig) {
        ROS_WARN("%s Cannot begin exploration until configuration parameters loaded", LOG_START);
        result.success = false;
    } else if (goal->mode == crosbot_explore::ExploreMode::MODE_LEFT_WALL) {
        Lock planningLock(planningThreadMutex);
        Lock drivingLock(drivingThreadMutex);

        searchParams.resetWallFollowData();
        searchParams.resetWaypointData();
        searchParams.side = SearchParameters::Left;
        searchParams.strategy = SearchParameters::WallFollow;

        result.success = true;
        voronoiConstraints = localVoronoiConstraints;
    } else if (goal->mode == crosbot_explore::ExploreMode::MODE_RIGHT_WALL) {
        Lock planningLock(planningThreadMutex);
        Lock drivingLock(drivingThreadMutex);

        searchParams.resetWallFollowData();
        searchParams.resetWaypointData();
        searchParams.side = SearchParameters::Right;
        searchParams.strategy = SearchParameters::WallFollow;

        result.success = true;
        voronoiConstraints = localVoronoiConstraints;
    } else if (goal->mode == crosbot_explore::ExploreMode::MODE_WAYPOINT) {
        Lock planningLock(planningThreadMutex);
        Lock drivingLock(drivingThreadMutex);


        searchParams.resetWallFollowData();
        searchParams.resetWaypointData();
        searchParams.strategy = SearchParameters::Waypoint;
        searchParams.targetOrientation = goal->targetOrientation;

        nav_msgs::Path path = goal->path;
        bool setResult = setWaypoints(path);

        result.success = setResult;
        voronoiConstraints = localVoronoiConstraints;
    } else if (goal->mode == crosbot_explore::ExploreMode::MODE_RESUME) {
        enableDrive = true;

        result.success = false;
    } else if (goal->mode == crosbot_explore::ExploreMode::MODE_STOP) {
        // On pause, just stop
        result.success = true;
    }
}

void ExplorerROSNode::configure() {
    //Lock lock(mutexReconfig);

    // read configuration/parameters
    ros::NodeHandle paramNH("~");
    paramNH.param<std::string>("base_frame", baseFrame, DEFAULT_BASEFRAME);

    ConfigElementPtr cfg = new ROSConfigElement(paramNH);
    debugMsgs = cfg->getParamAsBool("debug", false);
    gridSub_name = cfg->getParam("map_sub", "map");
    historySub_name = cfg->getParam("history_sub", "history");
    imagePub_name = cfg->getParam("image_pub", "/crosbot_explore/image");
    velPub_name = cfg->getParam("vel_pub", "cmd_vel");
    ConfigElementPtr voronoiCfg = cfg->getChild("voronoi");
    if (voronoiCfg != NULL) {
        localVoronoiConstraints.restricted = voronoiCfg->getParamAsDouble("restrict", localVoronoiConstraints.restricted);
        localVoronoiConstraints.partial = voronoiCfg->getParamAsDouble("partial", localVoronoiConstraints.partial);
        localVoronoiConstraints.expand = voronoiCfg->getParamAsDouble("expand", localVoronoiConstraints.expand);
        localVoronoiConstraints.orphanThreshold = voronoiCfg->getParamAsInt("orphan", localVoronoiConstraints.orphanThreshold);
        ROS_INFO("%s restrict: %.2lf - partial: %.2lf - expand: %.2lf - orphan: %d", LOG_START, localVoronoiConstraints.restricted, localVoronoiConstraints.partial, localVoronoiConstraints.expand, localVoronoiConstraints.orphanThreshold);
        voronoiConstraints = localVoronoiConstraints;
    }
    ConfigElementPtr searchCfg = cfg->getChild("search");
    if (searchCfg != NULL) {
        searchParams.minTravelDistance = searchCfg->getParamAsDouble("min_travel_distance", searchParams.minTravelDistance);
        searchParams.obstructionAngle = DEG2RAD(searchCfg->getParamAsDouble("obstruction_angle", RAD2DEG(searchParams.obstructionAngle)));
        searchParams.proximityAngle = DEG2RAD(searchCfg->getParamAsDouble("proximity_angle", RAD2DEG(searchParams.proximityAngle)));
        searchParams.proximityDistance = searchCfg->getParamAsDouble("proximity_distance", searchParams.proximityDistance);
        searchParams.skeletonSearchDistance = searchCfg->getParamAsDouble("skeleton_search_distance", searchParams.skeletonSearchDistance);
        searchParams.wallSearchDistance = searchCfg->getParamAsDouble("wall_search_distance", searchParams.wallSearchDistance);
        searchParams.wallSearchAngle = searchCfg->getParamAsDouble("wall_search_angle", searchParams.wallSearchAngle);
    }
    if (cfg != NULL) {
        driveParams.maxVel = cfg->getParamAsDouble("maxVel", driveParams.maxVel);
        driveParams.maxTurn = cfg->getParamAsDouble("maxTurn", driveParams.maxTurn);
        driveParams.minVel = cfg->getParamAsDouble("minVel", driveParams.minVel);
        driveParams.minTurn = cfg->getParamAsDouble("minTurn", driveParams.minTurn);
        ROS_INFO("%s maxVel: %.2lf, maxTurn: %.2lf", LOG_START, driveParams.maxVel, driveParams.maxTurn);
    }
    ConfigElementPtr hysteresisCfg = cfg->getChild("hysteresis");
    if (hysteresisCfg != NULL) {
        hysteresisParams.clusterTimeout = crosbot::Duration(hysteresisCfg->getParamAsDouble("cluster_timeout", hysteresisParams.clusterTimeout.toSec()));
        hysteresisParams.enabled = hysteresisCfg->getParamAsBool("enabled", hysteresisParams.enabled);
        hysteresisParams.hysteresisSize = hysteresisCfg->getParamAsInt("hysteresis_size", hysteresisParams.hysteresisSize);
        hysteresisParams.hysteresisTimeout = crosbot::Duration(hysteresisCfg->getParamAsDouble("hysteresis_timeout", hysteresisParams.hysteresisTimeout.toSec()));
        hysteresisParams.marginDistance = hysteresisCfg->getParamAsDouble("margin_distance", hysteresisParams.marginDistance);
        hysteresisParams.marginHeading = hysteresisCfg->getParamAsDouble("margin_heading", hysteresisParams.marginHeading);
    }

    ros::NodeHandle nh;
    subscribeGrid(nh);
    historySub = nh.subscribe(historySub_name, 1, &ExplorerROSNode::callbackHistory, this);
    imagePub = nh.advertise<sensor_msgs::Image>(imagePub_name, 1);
    velPub = nh.advertise<geometry_msgs::Twist>(velPub_name, 1);

    // Dynamic reconfiguration
    gotReconfig = false;
    reconfig_server = new ReconfigServer(mutexReconfig);
    publishReconfig(localVoronoiConstraints);
    reconfig_callbackType = boost::bind(&ExplorerROSNode::callback_reconfigure, this, _1, _2);
    reconfig_server->setCallback(reconfig_callbackType);

    // Configure simple action server and register callbacks and start
    actionServer_name = cfg->getParam("action_server_topic", "/crosbot/explorer");
    actionServer = new ExploreAServer(nh, actionServer_name, boost::bind(&ExplorerROSNode::callback_explore, this, _1), false);
    actionServer->start();

    ROS_INFO("%s configure done", LOG_START);
}

void ExplorerROSNode::shutdown() {
    Explorer::shutdown();

    // Stop action server
    actionServer->shutdown();
    delete actionServer;
    reconfig_server->clearCallback();
    gotReconfig = false;
    delete reconfig_server;

    // Shutdown subs/pubs
    gridSub.shutdown();
    historySub.shutdown();
    imagePub.shutdown();
    velPub.shutdown();
}

void ExplorerROSNode::subscribeGrid(ros::NodeHandle& nh) {
    gridSub.shutdown();

    ROS_INFO("%s Subscribing to map topic: %s", LOG_START, gridSub_name.c_str());
    gridSub = nh.subscribe(gridSub_name, 1, &ExplorerROSNode::callbackOccGrid, this);
}

VoronoiGridPtr ExplorerROSNode::getLatestVoronoi() {
    nav_msgs::OccupancyGridConstPtr latestMap;
    {
        Lock lock(rosLock, false);
        latestMap = this->latestMap;
    }

    if (latestMap == NULL) {
        return NULL;
    }
    if (latestVoronoi != NULL && latestVoronoi->timestamp == Time(latestMap->header.stamp)) {
        return latestVoronoi;
    }

    if (debugMsgs) {
        ROS_INFO("%s Loading new voronoi grid: restrict: %.1lf, partial: %.1lf, expand: %.1lf", LOG_START, voronoiConstraints.restricted, voronoiConstraints.partial, voronoiConstraints.expand);
        ROS_INFO("%s Time map: %.5lf, time now: %.5lf", LOG_START, latestMap->header.stamp.toSec(), ros::Time::now().toSec());
    }
    Pose robot = getLatestPose(latestMap->header.frame_id, latestMap->header.stamp);
    latestVoronoi = new VoronoiGrid(*latestMap, voronoiConstraints, robot); // TODO: include history
    return latestVoronoi;
}

Pose ExplorerROSNode::getLatestPose(std::string frame_id, Time timestamp) {
    //if (debugMsgs) ROS_INFO("getLatestPose() - %s, %s at time: %.5lf", frame_id.c_str(), baseFrame.c_str(), timestamp.toSec());

    tf::StampedTransform transform;
    try {
        tfListener.waitForTransform(frame_id, baseFrame, timestamp.toROS(), ros::Duration(1.0));
        tfListener.lookupTransform(frame_id, baseFrame, timestamp.toROS(), transform);
    } catch (tf::TransformException& e) {
        ROS_WARN("%s getLatestPose(): Exception caught(%s). time: %.5lf, %s - %s\n", LOG_START, e.what(), timestamp.toSec(), frame_id.c_str(), baseFrame.c_str());
        return Pose(INFINITY, INFINITY, INFINITY);
    }
    return Pose(transform);
}

Pose ExplorerROSNode::getLatestPose() {
    nav_msgs::OccupancyGridConstPtr latestMap;
    {
        Lock lock(rosLock, false);
        latestMap = this->latestMap;
    }

    if (latestMap == NULL) {
        return Pose(INFINITY, INFINITY, INFINITY);
    }
    return getLatestPose(latestMap->header.frame_id, latestMap->header.stamp);
}

Pose ExplorerROSNode::findDriveTarget(const VoronoiGrid& voronoi, const Pose& robot) {
    Pose rval = Explorer::findDriveTarget(voronoi, robot);

    // TODO: to view in camera relative frame need camera_info topic (see http://wiki.ros.org/rviz/DisplayTypes/Camera)
    if (imagePub.getNumSubscribers() > 0) {
        ImagePtr image = getPlanImage(voronoi, robot, rval);
        imagePub.publish(image->toROS());
    }

    return rval;
}

bool ExplorerROSNode::stopMotors() {
    if (((void*)(velPub)) == NULL) {
        return false;
    }

    ROS_INFO("%s stopMotors()", LOG_START);
    geometry_msgs::Twist twist;

    twist.linear.x = twist.linear.y = twist.linear.z = 0;
    twist.angular.x = twist.angular.y = twist.angular.z = 0;

    velPub.publish(twist);

    return true;
}

void ExplorerROSNode::driveTo(const Pose& relativePosition) {
    if (((void*)(velPub)) == NULL) {
        return;
    }


    double d = 0;   // Distance to travel
    double a = 0;   // Angle to rotate

    if (searchParams.rotateToPose) {
        double yaw, pitch, roll;
        relativePosition.getYPR(yaw, pitch, roll);
        d = 0;
        a = yaw;
        //ROS_INFO("%s rotate only relative yaw: %.2lf", LOG_START, RAD2DEG(yaw));
    } else {
        d = relativePosition.position.distanceTo(Point());
        a = atan2(relativePosition.position.y, relativePosition.position.x);
    }


    geometry_msgs::Twist twist;

    twist.linear.x = twist.linear.y = twist.linear.z = 0;
    twist.angular.x = twist.angular.y = twist.angular.z = 0;

    if (fabs(a) >= driveParams.turnOnly) {
        twist.angular.z = (a<0) ? -driveParams.maxTurn : driveParams.maxTurn;
    } else {
        twist.angular.z = a / driveParams.turnOnly * driveParams.maxTurn;
        twist.angular.z = std::min(twist.angular.z, driveParams.maxTurn);
        twist.angular.z = std::max(twist.angular.z, -driveParams.maxTurn);
        if (twist.angular.z < driveParams.minTurn && twist.angular.z > -driveParams.minTurn) {
            if (twist.angular.z > 0) {
                twist.angular.z = driveParams.minTurn;
            } else {
                twist.angular.z = -driveParams.minTurn;
            }
        }

        // The closer the point is to the robot, the slower the robot drives
        twist.linear.x = (d / searchParams.wallSearchDistance) * driveParams.maxVel * (1 - (fabs(a) / driveParams.turnOnly));
        twist.linear.x = std::min(twist.linear.x, driveParams.maxVel);
        twist.linear.x = std::max(twist.linear.x, -driveParams.maxVel);
        if (twist.linear.x < driveParams.minVel && twist.linear.x > -driveParams.minVel) {
            if (twist.linear.x > 0) {
                twist.linear.x = driveParams.minVel;
            } else {
                twist.linear.x = -driveParams.minVel;
            }
        }
        //ROS_INFO("%s twist command: linear.x: %.4lf, angular.z: %.4lf", LOG_START, twist.linear.x, twist.angular.z);
    }
    velPub.publish(twist);
}

bool ExplorerROSNode::setWaypoints(nav_msgs::Path& navPath) {
    // Get voronoi map
    VoronoiGridPtr voronoi = latestVoronoi;
    if (voronoi == NULL) {
        ROS_ERROR("%s No voronoi map to plot robot path on", LOG_START);
        return false;
    }

    // Transform map and nav path into the same frame
    // TODO: use stamped pose of each point in the path, rather than the frame id?
    tf::StampedTransform transform;
    try {
        ros::Time now = ros::Time::now();
        tfListener.waitForTransform(voronoi->frame, navPath.header.frame_id, now, ros::Duration(DEFAULT_MAXWAIT4TRANSFORM));
        tfListener.lookupTransform(voronoi->frame, navPath.header.frame_id, now, transform);
    } catch (tf::TransformException& ex) {
        ROS_ERROR("%s: Error getting transform for voronoi frame to navPath. (%s)\n", LOG_START, ex.what());
        return false;
    }

    // Set waypoints for search strategy
    std::vector< Pose > waypoints;
    waypoints.resize(navPath.poses.size());
    for (size_t i = 0; i < waypoints.size(); ++i) {
        waypoints[i] = transform * Pose(navPath.poses[i]).toTF();
    }
    searchParams.setWaypoints(waypoints, getLatestPose());

    return true;
}

void ExplorerROSNode::publishReconfig(VoronoiGrid::Constraints& constraints) {
    // Set config values
    crosbot_explore::CrosbotExploreReconfigConfig config;

    config.voronoi_restricted = localVoronoiConstraints.restricted;
    config.voronoi_partial = localVoronoiConstraints.partial;
    config.voronoi_expand = localVoronoiConstraints.expand;
    config.voronoi_orphanThreshold = localVoronoiConstraints.orphanThreshold;

    config.min_travel_distance = searchParams.minTravelDistance;
    config.map_sub = gridSub_name;
    config.obstruction_angle = RAD2DEG(searchParams.obstructionAngle);
    config.proximity_angle = RAD2DEG(searchParams.proximityAngle);
    config.proximity_distance = searchParams.proximityDistance;
    config.skeleton_search_distance = searchParams.skeletonSearchDistance;
    config.wall_search_distance = searchParams.wallSearchDistance;
    config.wall_search_angle = searchParams.wallSearchAngle;

    config.drive_max_turn = driveParams.maxTurn;
    config.drive_max_vel = driveParams.maxVel;

    // Call service
    reconfig_server->updateConfig(config);
}

void ExplorerROSNode::statusChanged(const std::string& status) {
    crosbot_explore::ExploreFeedback feedback;
    feedback.id = currentGoalId;
    feedback.status = status;
    feedback.operating = !paused;

    // TODO: convert string to use crosbot_explore::ExploreAction strings
    // TODO: Deal with "blocked" and "arrived" changed status

    if (searchParams.strategy == SearchParameters::WallFollow && searchParams.side == SearchParameters::Left) {
        feedback.searchStrategy = crosbot_explore::ExploreMode::MODE_LEFT_WALL;
    } else if (searchParams.strategy == SearchParameters::WallFollow && searchParams.side == SearchParameters::Right) {
        feedback.searchStrategy = crosbot_explore::ExploreMode::MODE_RIGHT_WALL;
    } else if (searchParams.strategy == SearchParameters::Waypoint) {
        feedback.searchStrategy = crosbot_explore::ExploreMode::MODE_WAYPOINT;
    }
    actionServer->publishFeedback(feedback);
}

} // namespace crosbot

