/*
 * move.cpp
 *
 *  Created on: 27/08/2014
 *      Author: mmcgill
 */

#include <ros/ros.h>

#include <crosbot/config.hpp>
#include <crosbot_explore/nodes/astarExplorerNode.hpp>

#define DEFAULT_PLANNING_RATE       1
#define LOG_START                   "MoveNode ::"
#define LOG_START_PLANNING          "MoveNodePlanningThread ::"

namespace crosbot {

PlanningThreadMove::PlanningThreadMove(MoveNode *moveNode) :
    crosbot::Thread("PlanningThreadMove"),
    crosbot::HandledObject(),
    moveNode(moveNode)
{
    if (moveNode == NULL){
        ROS_ERROR("%s Null move node", LOG_START_PLANNING);
    }
}

PlanningThreadMove::~PlanningThreadMove() {
}

void PlanningThreadMove::run() {
    if (moveNode != NULL) {
        moveNode->planningLoop();
    }
}

MoveNode::MoveNode() :
    planningOperating(false),
    replanRate(DEFAULT_PLANNING_RATE),
    aStarReconfig_server(NULL),
    differentPlanningGrid(false)
{
}

MoveNode::~MoveNode() {
}

void MoveNode::configure() {
    ExplorerROSNode::configure();

    ros::NodeHandle paramNH("~");
    paramNH.param<std::string>("astar_image_pub", astarImgPub_name, "/crosbot_explore/astar_voronoi_image");
    paramNH.param<double>("replanRate", replanRate, DEFAULT_PLANNING_RATE);

    // Allow different voronoi constraints for a* and other methods
    ConfigElementPtr cfg = new ROSConfigElement(paramNH);
    ConfigElementPtr astarCfg = cfg->getChild("astar");
    ConfigElementPtr voronoiCfg = astarCfg->getChild("voronoi");
    if (voronoiCfg != NULL) {
        astarVoronoiConstraints.restricted = voronoiCfg->getParamAsDouble("restrict", astarVoronoiConstraints.restricted);
        astarVoronoiConstraints.partial = voronoiCfg->getParamAsDouble("partial", astarVoronoiConstraints.partial);
        astarVoronoiConstraints.expand = voronoiCfg->getParamAsDouble("expand", astarVoronoiConstraints.expand);
        astarVoronoiConstraints.orphanThreshold = voronoiCfg->getParamAsInt("orphan", astarVoronoiConstraints.orphanThreshold);
        ROS_INFO("%s restrict: %.2lf - partial: %.2lf - expand: %.2lf - orphan: %d", LOG_START, astarVoronoiConstraints.restricted, astarVoronoiConstraints.partial, astarVoronoiConstraints.expand, astarVoronoiConstraints.orphanThreshold);
        voronoiConstraints = astarVoronoiConstraints;
    }

    // Configure A* Planner
    ConfigElementPtr astarPlannerCfg = astarCfg->getChild("planner");
    astarGridSub_name = astarCfg->getParam("map_sub", gridSub_name);
    differentPlanningGrid = astarGridSub_name != gridSub_name;
    AStarParameters astarParams = planner.getParameters();
    astarParams.k = astarCfg->getParamAsDouble("risk_k", astarParams.k);
    astarParams.weight_expanded = astarCfg->getParamAsDouble("weight_expanded", astarParams.weight_expanded);
    astarParams.weight_partial = astarCfg->getParamAsDouble("weight_partial", astarParams.weight_partial);
    planner.setParamaters(astarParams);

    // Dynamic reconfiguration
    ros::NodeHandle reconfigNh("~/astar");
    aStarReconfig_server = new AStarReconfigServer(aStarMutexReconfig, reconfigNh);
    publishReconfig(astarParams);
    aStarReconfig_callbackType = boost::bind(&MoveNode::callback_astarReconfigure, this, _1, _2);
    aStarReconfig_server->setCallback(aStarReconfig_callbackType);

    // Publishers/Subscribers
    ros::NodeHandle nh;
    astarImagePub = nh.advertise<sensor_msgs::Image>(astarImgPub_name, 1);
    if (differentPlanningGrid) {
        astarGridSub = nh.subscribe(astarGridSub_name, 1, &MoveNode::callback_astarOccGrid, this);
    }

    // Planning Thread
    planningThread = new PlanningThreadMove(this);
}

void MoveNode::shutdown() {
    ExplorerROSNode::shutdown();

    // Close publishers/subscribers
    astarGridSub.shutdown();

    // Ensure planning thread has stopped
    planningOperating = false;
    for (uint32_t i = 0; i < DEFAULT_WAIT_FOR_THREAD_CLOSE && planningThread->isAlive(); i + 100) {
        usleep(100000);
    }
    if (planningThread->isAlive()) {
        ROS_ERROR("%s Planning thread failed to shut down.", LOG_START);
    }
}

void MoveNode::callback_astarOccGrid(const nav_msgs::OccupancyGridConstPtr& astarLatestMap) {
    Lock lock(astarRosLock, true);
    this->astarLatestMap = astarLatestMap;
}

void MoveNode::callback_astarReconfigure(crosbot_explore::CrosbotAStarReconfigConfig &config, uint32_t level) {
    astarVoronoiConstraints.restricted = config.astar_voronoi_restricted;
    astarVoronoiConstraints.partial = config.astar_voronoi_partial;
    astarVoronoiConstraints.expand = config.astar_voronoi_expand;
    astarVoronoiConstraints.orphanThreshold = config.astar_voronoi_orphanThreshold;
    ROS_INFO("%s new dynamic config: \t restrict: %.2lf, partial %.2lf, expand: %.2lf", LOG_START, astarVoronoiConstraints.restricted, astarVoronoiConstraints.partial, astarVoronoiConstraints.expand);

    AStarParameters astarParams = planner.getParameters();
    astarParams.weight_expanded = config.weight_expanded;
    astarParams.weight_partial = config.weight_partial;
    planner.setParamaters(astarParams);
    ROS_INFO("%s Got A* reconfig. W(expand): %.2lf, W(partial):%.2lf", LOG_START, config.weight_expanded, config.weight_partial);
}

void MoveNode::publishReconfig(AStarParameters& params) {
    // Set config values
    crosbot_explore::CrosbotAStarReconfigConfig config;
    config.astar_voronoi_expand = astarVoronoiConstraints.expand;
    config.astar_voronoi_orphanThreshold = astarVoronoiConstraints.orphanThreshold;
    config.astar_voronoi_partial = astarVoronoiConstraints.partial;
    config.astar_voronoi_restricted = astarVoronoiConstraints.restricted;
    config.weight_expanded = params.weight_expanded;
    config.weight_partial = params.weight_partial;

    // Call service
    aStarReconfig_server->updateConfig(config);
}

VoronoiGridPtr MoveNode::getLatestAstarVoronoi() {
    nav_msgs::OccupancyGridConstPtr latestMap;

    if (differentPlanningGrid) {
        {
            Lock lock(astarRosLock, false);
            latestMap = this->astarLatestMap;
        }

        if (latestMap == NULL) {
            return NULL;
        }
        if (astarLatestVoronoi != NULL && astarLatestVoronoi->timestamp == Time(latestMap->header.stamp)) {
            return astarLatestVoronoi;
        }

        if (debugMsgs) {
            ROS_INFO("%s Loading new A* voronoi grid: restrict: %.1lf, partial: %.1lf, expand: %.1lf", LOG_START, astarVoronoiConstraints.restricted, astarVoronoiConstraints.partial, astarVoronoiConstraints.expand);
            ROS_INFO("%s Time map: %.5lf, time now: %.5lf", LOG_START, latestMap->header.stamp.toSec(), ros::Time::now().toSec());
        }
        Pose robot = getLatestPose(latestMap->header.frame_id, Time::now());
        astarLatestVoronoi = new VoronoiGrid(*latestMap, astarVoronoiConstraints, robot);

    } else {
        astarLatestVoronoi = getLatestVoronoi();
    }

    return astarLatestVoronoi;
}

void MoveNode::handleExploreMode(const crosbot_explore::ExploreGoalConstPtr& goal,
        crosbot_explore::ExploreResult& result, bool& enableDrive)  {
    if (goal->mode == crosbot_explore::MoveMode::MODE_ASTAR) {
        ROS_INFO("%s A* mode activated", LOG_START);
        enableDrive = false;

        // A* goal is first element of path
        if (goal->path.poses.size() > 0) {
            searchParams.resetWallFollowData();
            searchParams.resetWaypointData();
            searchParams.strategy = SearchParameters::Waypoint;
            searchParams.targetOrientation = goal->targetOrientation;
            voronoiConstraints = astarVoronoiConstraints;

            // TODO: go through path of goals - so you can have sequence of A* goals
            Pose3DStamped astarGoal(goal->path.poses.front());
            setCurrentAStarGoal(astarGoal);

            // Set first waypoints
            VoronoiGridPtr voronoi;
            bool setResult = updateWaypoints();
            result.success = setResult;

            // Start planning thread
            planningOperating = true;
            planningThread->start();
        } else {
            ROS_ERROR("%s No pose provided for A* goal", LOG_START);
            result.success = false;
        }

    } else {
        ExplorerROSNode::handleExploreMode(goal, result, enableDrive);

        // If mode has changed, stop planning thread
        if (!(goal->mode == crosbot_explore::MoveMode::MODE_RESUME || goal->mode == crosbot_explore::MoveMode::MODE_STOP)) {
            planningOperating = false;
        }
    }
}

void MoveNode::planningLoop() {
    ros::Rate rate(replanRate);

    VoronoiGridPtr voronoi;

    while (planningOperating) {
        if (!paused) {
            if (searchParams.atDestination) {
                // Check if at goal, pause exploration
                pause();
            } else {
                // Otherwise, update waypoints and continue driving
                bool updated = updateWaypoints();

                // If updated failed - pause
                if (!updated) {
                    pause();
                }
            }
        }

        // Sleep even if update failed to allow pause to go through first
        rate.sleep();
    }
}

bool MoveNode::setWaypoints(Plan& plan, const std::string& planFrameId, const std::string& navPathFrameId, const crosbot::Time time) {
    bool retVal = false;

    // Establish converted plan
    bool converted = false;
    std::string convertedFrameId;
    Plan emptyPlan;
    Plan revisedPlan;
    Plan& convertedPlan = emptyPlan;
    if (differentPlanningGrid) {
        // Convert plan points into frame of reference of navigation voronoi grid
        try {
            // Ensure transform
            //ROS_INFO("%s transforming from frame: %s to frame: %s", LOG_START, planFrameId.c_str(), navPathFrameId.c_str());
            tfListener.waitForTransform(navPathFrameId, planFrameId, time.toROS(), ros::Duration(0.5));
            //ROS_INFO("%s \t got transform", LOG_START);

            // Generate new path
            for (Pose& pose : plan) {
                geometry_msgs::PoseStamped oldPose;
                oldPose.header.frame_id = planFrameId;
                oldPose.header.stamp = time.toROS();
                oldPose.pose = pose.toROS();
                geometry_msgs::PoseStamped newPose;
                newPose.header.frame_id = navPathFrameId;
                tfListener.transformPose(navPathFrameId, oldPose, newPose);
                revisedPlan.push_back(crosbot::Pose(newPose));
            }

            convertedPlan = revisedPlan;
            convertedFrameId = navPathFrameId;
            converted = true;
        } catch (tf::TransformException& e) {
            ROS_WARN("%s setWaypoints(): Exception caught(%s).\n", LOG_START, e.what());
            converted = false;
        }
    } else {
        convertedPlan = plan;
        convertedFrameId = planFrameId;
        converted = true;
    }

    // Generate navPath
    if (converted) {
        nav_msgs::Path navPath;
        navPath.header.frame_id = convertedFrameId;
        navPath.header.stamp = time.toROS();
        navPath.poses.resize(convertedPlan.size());
        for (size_t i = 0; i < convertedPlan.size(); ++i) {
            navPath.poses[i].pose = convertedPlan[i].toROS();
        }
        retVal = ExplorerROSNode::setWaypoints(navPath);
    } else {
        ROS_WARN("%s failed to convert path to navigation frame", LOG_START);
    }


    return retVal;
}

void MoveNode::setCurrentAStarGoal(const Pose3DStamped& goal) {
    Lock lock(dataLock, true);
    currentAStarGoal = goal;
    ROS_INFO("%s Setting current A* Goal at: %s in frame %s", LOG_START, currentAStarGoal.toString().c_str(), goal.frame_id.c_str());
}

Pose3DStamped MoveNode::getCurrentAStarGoal(std::string frame_id, Time timestamp) {
    // Does not take lock as calling code already has lock

    Pose3DStamped aStarGoal(INFINITY, INFINITY, INFINITY, frame_id, Time());
    if (currentAStarGoal.isFinite()) {
        tf::StampedTransform transform;
        try {
            // Change time of goal to now
            geometry_msgs::PoseStamped goal = currentAStarGoal.toROS();
            goal.header.stamp = timestamp.toROS();

            geometry_msgs::PoseStamped aStarOut;
            tfListener.waitForTransform(frame_id, goal.header.frame_id, timestamp.toROS(), ros::Duration(0.5));
            tfListener.transformPose(frame_id, goal, aStarOut);
            aStarGoal = aStarOut;
        } catch (tf::TransformException& e) {
            ROS_WARN("%s getCurrentAStarGoal(): Exception caught(%s).\n", LOG_START, e.what());
        }
    }

    return aStarGoal;
}

bool MoveNode::updateWaypoints() {
    //ROS_INFO("%s Updating Waypoints", LOG_START);

    // Load latest data to plan with
    Pose goal;
    Pose currentPose(INFINITY, INFINITY, INFINITY);
    VoronoiGridPtr astarVoronoi = NULL;
    VoronoiGridPtr navVoronoi = NULL;
    nav_msgs::OccupancyGridConstPtr currentMap;
    Time tNow(ros::Time::now());
    {{
        Lock lock(dataLock);
        bool loaded = true;
        std::string error;

        // Get latest voronoi
        astarVoronoi = getLatestAstarVoronoi();
        navVoronoi = getLatestVoronoi();
        if (astarVoronoi == NULL || navVoronoi == NULL) {
            loaded = false;
            error = "Voronoi grid(s) not available";
        }
        // If the grid takes a long time to compute, then reset time
        tNow = Time(ros::Time::now());

        // Get current pose of the robot (in frame of reference of voronoi)
        if (loaded) {
            if (debugMsgs) ROS_INFO("%s getting latest pose at: %.5lf", LOG_START, tNow.toSec());
            currentPose = getLatestPose(astarVoronoi->frame, tNow);
            if (!currentPose.isFinite()) {
                loaded = false;
                error = "Unable to determine current pose";
            }
        }

        // Set current A* goal
        if (loaded) {
            goal = getCurrentAStarGoal(astarVoronoi->frame, tNow);
            if (!goal.isFinite()) {
                loaded = false;
                error = "Goal is infinite";
            }
        }

        if (!loaded) {
            ROS_WARN("%s Cannot update waypoints: %s", LOG_START, error.c_str());
            return false;
        }
    }}
    //ROS_INFO("%s Updating Waypoints: loaded data", LOG_START);

    // Get A* voronoi grid image to load debug images
    ImagePtr debugImage;
    if (astarImagePub.getNumSubscribers() > 0) {
        debugImage = astarVoronoi->getImage();
    }

    // Plan path to goal
    bool waypointsSet = false;
    Plan plan = planner.getPath(astarVoronoi, currentPose, goal, debugImage);
    if (debugMsgs) ROS_INFO("%s Loaded path with %d waypoints to %s", LOG_START, (int) plan.size(), goal.position.toString().c_str());
    if (plan.size() > 0) {
        // send path to explorer
        waypointsSet = setWaypoints(plan, astarVoronoi->frame, navVoronoi->frame, tNow);
        //ROS_INFO("%s Waypoints set", LOG_START);
    } else {
        ROS_WARN("%s Unable to plan a path to %.3lf, %.3lf, %.3lf.", LOG_START, goal.position.x, goal.position.y, goal.position.z);
    }

    // Publish A* voronoi grid
    if (astarImagePub.getNumSubscribers() > 0 && debugImage != NULL) {
        astarImagePub.publish(debugImage->toROS());
    }

    return waypointsSet;
}

} // namespace crosbot


int main(int argc, char** argv) {
    ros::init(argc, argv, "move");

    crosbot::MoveNode node;
    node.configure();

    while (ros::ok()) {
        ros::spin();
    }
    node.shutdown();

    return 0;
}
