/*
 * fastslam.hpp
 *
 *  Created on: 20/02/2012
 *      Author: rescue
 */

#ifndef CROSBOT_FASTSLAM_HPP_
#define CROSBOT_FASTSLAM_HPP_

#include <crosbot_fastslam/particle.hpp>
#include <crosbot_map/map.hpp>
#include <crosbot/thread.hpp>

namespace crosbot {

namespace fastslam {

class FastSLAMMap;
typedef Handle<FastSLAMMap> FastSLAMMapPtr;

class FastSLAMMap : public Map {
public:
	FastSLAMMap();
	virtual ~FastSLAMMap();

	/**
	 * Methods for configuring and starting/stopping.
	 */
	void configure(ConfigElementPtr config);
	void start();
	void stop();

	/**
	 * Allow for map to be paused resumed.
	 */
	void setPaused(bool pause);
	bool isPaused();

	/**
	 * Allow map to be set in track only mode.
	 */
	void setUpdateMap(bool update);

	/**
	 * Map can be cloned, saved and loaded.
	 */
	MapPtr clone();
	bool load(std::string filename) throw(IOException);
	bool save(std::string filename) throw(IOException);

	/**
	 * Gets the current pose and position tracker pose.
	 */
	Pose getCurrentPose(Pose* trackerPose = NULL);

	/**
	 * Gets the tags in the map.
	 */
	TagListPtr getTags();

	/**
	 * Gets the path the robot has followed if the map allows it.
	 */
	PathPtr getPath();

	/**
	 * Allows the robots pose to be reset.
	 */
	void resetPose(Pose3D pose = Pose());

	/**
	 * Data for maps. Cloud should be relative to the robot.
	 */
	void newPointCloud(PointCloudPtr cloud, Pose odometry, Pose sensor);

	/**
	 * Add new tag to the map
	 */
	void addTag(TagPtr tag);

	/**
	 * Convert to a simple map type.
	 */
	nav_msgs::OccupancyGridPtr asOccupancyGrid();
	nav_msgs::OccupancyGridPtr asOccupancyGrid(ParticlePtr);

protected:
	struct MotionModel {
	public:
		double theta, drift, range, rangeXY, shift, shiftXY, slip;
		double smearX, smearY, smearTheta;

		MotionModel() :
			theta(KTH), drift(KDR), range(KR), rangeXY(KRXY),
			shift(KSH), shiftXY(KSHXY), slip(KSLP),
			smearX(SMEAR_X), smearY(SMEAR_Y), smearTheta(SMEAR_YAW)
		{}
	};

	/**
	 * Actions to process in changing FastSLAM operation
	 */
    class QueuedAction {
    public:
        enum Type { Pause, Resume, ResetPose };
        Type type;
        Pose pose;
    };

    /**
     * Thread in which to run slam algorithm
     */
    class FastSLAMThread : public Thread {
    protected:
        FastSLAMMapPtr map;

    public:
        FastSLAMThread(FastSLAMMapPtr map) :
            Thread("FastSLAMMap::FastSLAMThread"),
            map(map) {
            setGuarded(false);
        }

        void run() {
            map->slam();
        }
    };



	/**
	 * FastSLAM motion parameters, and motion model
	 */
    FastSLAMParameters parameters;
    MotionModel motionModel;
    double motionSumXY;
    double motionSumTheta;

    /**
     * The mean and motion particles
     */
    ParticlePtr mean;
    ParticlePtr motion;

    /**
     * All of the particles
     */
    std::vector<ParticlePtr> particles;
    ReadWriteLock particlesLock;

    /**
     * List of poses of particles, used in rendering
     */
    std::vector<Pose> particlePoses;
    Pose meanPose;

    /**
     * Concurrency and SLAM processing operations
     */
	Mutex incomingMutex;
	MapCloudPtr latestPointCloud;
	std::queue<TagPtr> receivedTags;
	std::queue<QueuedAction> queuedActions;
	bool slamOperating;
	bool slamPaused;
	bool slamUpdating;
	Semaphore slamSemaphore;

	/**
	 * SLAM thread
	 */
	FastSLAMThread slamThread;

	/**
	 * SLAM update loop
	 */
	void slam();
	void insertTag(TagPtr tag);
	void moveAndUpdate(MapCloudPtr cloud);
	void moveParticles(Pose relativeMotion, MapCloudPtr cloud, bool calculateWeight = true);
	void updateParticles(MapCloudPtr cloud);
	void resample();
	void resampleAndUpdate(MapCloudPtr cloud);

	/**
	 * Friend classes
	 */
	friend class FastSLAMRender;
	friend class FastSLAMThread;

};

} // namespace fastslam

} //namespace crosbot


#endif /* CROSBOT_FASTSLAM_HPP_ */
