/*
 * fastslam.cpp
 *
 *  Created on: 20/02/2012
 *      Author: rescue
 */

#include <ros/ros.h>

#include <crosbot_fastslam/fastslam.hpp>
#include <crosbot/utils.hpp>
#include <crosbot_fastslam/serialization.hpp>

#define LOG_START       "FastSLAM ::"

namespace crosbot {

namespace fastslam {

FastSLAMMap::FastSLAMMap() :
	motionSumXY(0), motionSumTheta(0),
	slamOperating(true), slamPaused(false), slamUpdating(true),
	slamThread(this)
{
}

FastSLAMMap::~FastSLAMMap() {
	stop();
}

void FastSLAMMap::configure(ConfigElementPtr config) {
	ROS_INFO("%s configure()", LOG_START);

    parameters.gain = config->getParamAsDouble("gain", parameters.gain);
	parameters.numberOfParticles = config->getParamAsInt("particles", parameters.numberOfParticles);
	if (config->hasParam("initialOdds")) {
		//LOG("BEFORE: %lf\n", parameters.initialOccupiedOdds);
		parameters.initialOccupiedOdds =
				ProbabilityTable.lodds(config->getParamAsDouble("initialOdds", DEFAULT_UNOBSERVEDLODDS));
		//LOG("AFTER: %lf\n", parameters.initialOccupiedOdds);
	}
	parameters.initialZ = config->getParamAsDouble("z", parameters.initialZ);
	parameters.updateThresholdXY = config->getParamAsDouble("update", parameters.updateThresholdXY);
	parameters.updateThresholdTheta = DEG2RAD(config->getParamAsDouble("updateTheta", DEFAULT_ANGULARTHRESHOLD));
	parameters.minDistanceBetweenSnaps = config->getParamAsDouble("snapThreshold", parameters.minDistanceBetweenSnaps);

	parameters.minHeight = config->getParamAsDouble("min", parameters.minHeight);
	parameters.maxHeight = config->getParamAsDouble("max", parameters.maxHeight);

	if (config->hasParam("initialOdds")) {
		parameters.initialOccupiedOdds = ProbabilityTable.lodds(config->getParamAsDouble("initialOdds", DEFAULT_UNOBSERVEDLODDS));
	}

	ConfigElementPtr motionConfig = config->getChild("motion");
	if (motionConfig == NULL) {
		motionConfig = config;
	}
	motionModel.range = motionConfig->getParamAsDouble("range", motionModel.range);
	motionModel.rangeXY = motionConfig->getParamAsDouble("rangeXY", motionModel.rangeXY);
	motionModel.theta = motionConfig->getParamAsDouble("theta", motionModel.theta);
	motionModel.drift = motionConfig->getParamAsDouble("drift", motionModel.drift);
	motionModel.shift = motionConfig->getParamAsDouble("shift", motionModel.shift);
	motionModel.shiftXY = motionConfig->getParamAsDouble("shiftXY", motionModel.shiftXY);
	motionModel.slip = motionConfig->getParamAsDouble("slip", motionModel.slip);

	ConfigElementPtr smearConfig = config->getChild("smear");
	if (smearConfig == NULL) {
		smearConfig = motionConfig;
	}
	motionModel.smearX = smearConfig->getParamAsDouble("x", motionModel.smearX);
	motionModel.smearY = smearConfig->getParamAsDouble("y", motionModel.smearY);
	motionModel.smearTheta = smearConfig->getParamAsDouble("theta", motionModel.smearTheta);

	ConfigElementPtr searchConfig = config->getChild("search");
	if (searchConfig == NULL) {
		searchConfig = config;
	}
	parameters.searchDistance = searchConfig->getParamAsDouble("search", parameters.searchDistance);
	parameters.searchDistance = searchConfig->getParamAsDouble("distance", parameters.searchDistance);
	parameters.searchFOV = DEG2RAD(searchConfig->getParamAsDouble("fov", RAD2DEG(parameters.searchFOV)));
	if (searchConfig->hasParam("pose")) {
		parameters.searchPose = searchConfig->getParamAsPose("pose", parameters.searchPose);
	}

	parameters.addMotionCopy = config->getParamAsBool("add_motion_copy", parameters.addMotionCopy);

	//ROS_INFO("%s configure()", LOG_START);
}

void FastSLAMMap::start() {
	slamOperating = true;
	if (!slamThread.isAlive()) {
		slamThread.start();
	}
}

void FastSLAMMap::stop() {
	slamOperating = false;
	slamSemaphore.notify();

	for (unsigned int i = 0; i < DEFAULT_WAIT_FOR_THREAD_CLOSE && slamThread.isAlive(); i+=10) {
		usleep(10000);
	}

	if (slamThread.isAlive()) {
		ERROR("FastSLAMMap: SlamThread chose not to complete in %f seconds.\n", DEFAULT_WAIT_FOR_THREAD_CLOSE/1000.0);
	}
}

void FastSLAMMap::setPaused(bool pause) {
    QueuedAction action;
    action.type = pause ? QueuedAction::Pause : QueuedAction::Resume;

    Lock lock(incomingMutex);
    queuedActions.push(action);

    slamSemaphore.notify();
}

bool FastSLAMMap::isPaused() {
    return slamPaused;
}

void FastSLAMMap::setUpdateMap(bool update) {
    slamUpdating = update;
}

MapPtr FastSLAMMap::clone() {
	FastSLAMMapPtr rval = new FastSLAMMap();
	rval->parameters = parameters;
	rval->motionModel = motionModel;
	rval->start();

	return rval;
}

bool FastSLAMMap::load(std::string filename) throw (IOException) {
	setPaused(true);
	for (uint32_t i = 0; !isPaused() && i < 10000; i += 10) {
		usleep(10000);
	}

	if (!isPaused()) {
		ERROR("FastSLAM map is not pausing. Cannot load from file.\n");
		return false;
	}

	ParticlePtr newMean = new  Particle(parameters);
	serialization::Serializer<Particle> serializer;
	serialization::FileInputStream fis(filename);
	serializer.read(*(newMean), fis);

	{
	    Lock lock(particlesLock, true);

	    mean = newMean;
		meanPose = mean->pose;
		motion = new Particle(newMean);

		particles.clear();
		particles.push_back(mean);
		particlePoses.clear();
		particlePoses.push_back(meanPose);
	}

	mapChanged();

	return true;
}

bool FastSLAMMap::save(std::string filename) throw (IOException) {
	if (mean == NULL) {
		return false;
	}

	ParticlePtr particle = new Particle(mean);

	serialization::Serializer<Particle> serializer;
	serialization::FileOutputStream fos(filename);
	serializer.write(*(particle), fos);

	return true;
}

Pose FastSLAMMap::getCurrentPose(Pose* trackerPose) {
	ParticlePtr particle = mean;
	if (particle == NULL) {
		if (trackerPose != NULL) {
			*trackerPose = Pose(NAN, NAN, NAN);
		}
		return Pose();
	}

	{
	    Lock lock(particle->rwLock);

	    if (trackerPose != NULL) {
			*trackerPose = particle->trackerPose;
		}
		return particle->pose;
	}
}

Map::TagListPtr FastSLAMMap::getTags() {
	ParticlePtr particle = mean;
	if (particle == NULL) {
		return NULL;
	}
	return particle->tags;
}

PathPtr FastSLAMMap::getPath() {
	ParticlePtr particle = mean;
	if (particle == NULL) {
		return NULL;
	}
	return particle->getPath();
}

void FastSLAMMap::resetPose(Pose3D pose) {
    QueuedAction action;
    action.type = QueuedAction::ResetPose;
    action.pose = pose;

    Lock lock(incomingMutex);
    queuedActions.push(action);
    slamSemaphore.notify();
}

void FastSLAMMap::newPointCloud(PointCloudPtr cloud, Pose odometry, Pose sensor) {
    Lock lock(incomingMutex);
    latestPointCloud = new MapCloud(*cloud, odometry, sensor);
    slamSemaphore.notify();
}

void FastSLAMMap::addTag(TagPtr tag) {
    Lock lock(incomingMutex);
    receivedTags.push(tag);
    slamSemaphore.notify();
}

nav_msgs::OccupancyGridPtr FastSLAMMap::asOccupancyGrid() {
	// Get copy of mean
	ParticlePtr p;
	if (mean == NULL) {
		p = new Particle(parameters);
	} else {
		p = new Particle(mean);
	}

	return asOccupancyGrid(p);
}

nav_msgs::OccupancyGridPtr FastSLAMMap::asOccupancyGrid(ParticlePtr particle) {
	nav_msgs::OccupancyGridPtr rval(new nav_msgs::OccupancyGrid());
	rval->info.origin = Pose().toROS();

	// Get map dimensions and origin

	Lock lock(particle->rwLock);
	HeightMultiMapPtr hmm;
	if (particle != NULL) {
		hmm = particle->map;
	}

	if (particle == NULL || hmm == NULL) {
		rval->info.origin = Pose().toROS();
		rval->info.height = rval->info.width = 0;
		rval->info.resolution = 0;

		rval->data.clear();
		return rval;
	}

	unsigned int minHMX = hmm->columns, maxHMX = 0, minHMY = hmm->rows, maxHMY = 0;

	rval->info.origin.position.x = INFINITY; rval->info.origin.position.y = INFINITY;

	// find which HeightMaps have data and need to be drawn
	for (unsigned int i = 0; i < hmm->rows; i++) {
		for (unsigned int j = 0; j < hmm->columns; j++) {
			HeightMapPtr hm = hmm->patches[i][j];
			if (hm != NULL) {
				if (i < minHMY) {
					minHMY = i;
				}
				if (i > maxHMY) {
					maxHMY = i;
				}
				if (j < minHMX) {
					minHMX = j;
				}
				if (j > maxHMX) {
					maxHMX = j;
				}

				if (hm->xOrig < rval->info.origin.position.x) {
					rval->info.origin.position.x = hm->xOrig;
				}
				if (hm->yOrig < rval->info.origin.position.y) {
					rval->info.origin.position.y = hm->yOrig;
				}
			}
		}
	}

	if (minHMX > maxHMX || minHMY > maxHMY) {
		ROS_WARN("FastSLAMRender: No data in map to export to GeoTIFF.\n");
		rval->info.origin = Pose().toROS();
		return rval;
	}

	// create image
	unsigned int imageWidth = (maxHMX + 1 - minHMX) * hmm->parameters.patchColumns;
    unsigned int imageHeight = (maxHMY + 1 - minHMY) * hmm->parameters.patchRows;

	rval->info.height = imageHeight; rval->info.width = imageWidth;
	rval->info.resolution = CELL_WIDTH;
	rval->data.resize(imageHeight * imageWidth);

	for (unsigned int i = minHMY; i <= maxHMY; i++) {
		for (unsigned int j = minHMX; j <= maxHMX; j++) {
			HeightMapPtr hm = hmm->patches[i][j];
			int offX = (j - minHMX) * hmm->parameters.patchColumns;
			int offY = (i - minHMY) * hmm->parameters.patchRows;
			if (hm == NULL) {
				for (unsigned int il = 0; il < hmm->parameters.patchRows; il++) {
					for (unsigned int jl = 0; jl < hmm->parameters.patchColumns; jl++) {
						int pixel = (offY + il) * imageWidth + offX + jl;
						rval->data[pixel] = -1;
					}
				}
			} else {
				for (unsigned int il = 0; il < hmm->parameters.patchRows; il++) {
					for (unsigned int jl = 0; jl < hmm->parameters.patchColumns; jl++) {
						HeightVal hv = *(hm->getIndex(il * hmm->parameters.patchColumns + jl));

						int pixel = (offY + il) * imageWidth + offX + jl;

						if (hv.observations <= 0) {
							rval->data[pixel] = -1;
						} else {
							double conf = ProbabilityTable.invodds(hv.prob);
							rval->data[pixel] = conf * 100;
						}
					}
				}
			}
		}
	}

	return rval;
}

void FastSLAMMap::slam() {
    // TODO: Get rid of "continues" from the loop - I missed them the first time!
	while (slamOperating) {
	    // Wait on new data or event to process
		slamSemaphore.wait();

		// Latest point cloud to update with
		MapCloudPtr update = NULL;
		{
			Lock lock(incomingMutex);

			if (!slamPaused && receivedTags.size() > 0) {
			    // Add new tags

			    TagPtr tag = receivedTags.front();
				receivedTags.pop();
				lock.unlock();
				insertTag(tag);
				tagAdded(tag);

				slamSemaphore.notify();
			} else if (queuedActions.size() > 0) {
			    // Process queued action

				QueuedAction action = queuedActions.front();
				queuedActions.pop();
				lock.unlock();

				if (action.type == QueuedAction::Pause) {
					slamPaused = true;
				} else if (action.type == QueuedAction::Resume) {
					slamPaused = false;
				} else if (action.type == QueuedAction::ResetPose) {
					// Reset pose in FastSLAM map.
					Lock lock2(particlesLock, true);

					motion->setPose(action.pose);
					for (uint32_t i = 0; i < particles.size(); i++) {
						particles[i]->setPose(action.pose);
					}
					mapChanged();
				}

				slamSemaphore.notify();
			} else {
			    // Load and process latest point cloud
			    update = latestPointCloud;
			    latestPointCloud = NULL;
			}
		}

		// If paused or no update, ignore notify loop
		if (slamPaused || update == NULL) {
		} else {
		    moveAndUpdate(update);

            // Ensure re-process
            slamSemaphore.notify();
		}
	}

	// If the loop terminates - ensure slam is paused
	slamPaused = true;
}

void FastSLAMMap::insertTag(TagPtr tag) {
	if (mean == NULL) {
		// No mean particle so create one
		mean = new Particle(parameters);
		mean->addTag(tag, true);
		return;
	}

	mean->addTag(tag, true);
	motion->addTag(tag);
	Lock lock(particlesLock, true);
	for (size_t i = 0; i < particles.size(); i++) {
		if (particles[i] != mean) {
			particles[i]->addTag(tag);
		}
	}
}

void FastSLAMMap::moveAndUpdate(MapCloudPtr cloud) {
	if (motion == NULL) {
		// No motion particle means FastSLAM has not yet been started
	    // Generate the motion particle
		ParticlePtr newMean = mean, newMotion;
		if (newMean == NULL) {
			newMean = new Particle(parameters);
		}
		newMean->trackerPose = cloud->odometry;
		newMean->history.push_back(Particle::History(newMean->pose, cloud, true));
		newMean->poseTransform = newMean->pose.toTF();

		if (slamUpdating) {
			newMean->motionCloud = cloud;
			newMean->update(newMean->motionCloud);
		}

		newMotion = new Particle(newMean);
		{
		    Lock lock(particlesLock, true);

		    mean = newMean;
			motion = newMotion;
			particles.clear();
			particles.push_back(mean);
			particlePoses.push_back(mean->pose);
			meanPose = mean->pose;
		}

		resample();
		mapChanged();
	} else {
	    // Update the motion particle, then update fastslam map

        // Get motion
        double yaw;
        double pitch;
        double roll;
        motion->trackerPose.getYPR(yaw, pitch, roll);

        Pose motionPose;
        tf::Transform motionTransform;
        tf::Transform trackerTransform = cloud->odometry.toTF();

        motionTransform = motion->trackerPose.toTF().inverse() * trackerTransform;
        motionPose = motionTransform;

        motionPose.getYPR(yaw, pitch, roll);
        motionSumTheta += fabs(yaw);
        motionSumXY += sqrt(motionPose.position.x*motionPose.position.x + motionPose.position.y*motionPose.position.y);

        // Decide is resampling is required
        bool resampleNeeded = true;
        if (motionSumXY < parameters.updateThresholdXY && motionSumTheta < parameters.updateThresholdTheta) {
            resampleNeeded = false;
        }

        // Update particles with motion
        moveParticles(motionPose, cloud, resampleNeeded);

        if (!resampleNeeded) {
            // If resampling not required, then track motion
            motionTracked();
        } else {
            // If resampling, then update particles and map
            motionSumXY = motionSumTheta = 0.0;

            // Combine update and resample into one more efficient block
            if (slamUpdating) {
                //ROS_INFO("%s Updating map", LOG_START);
                resampleAndUpdate(cloud);
            } else {
                //ROS_INFO("%s Resampling only", LOG_START);
                resample();
            }

            mapChanged();
        }
	}
}


void FastSLAMMap::moveParticles(Pose relativeMotion, MapCloudPtr cloud, bool calculateWeight) {
	tf::Transform motionTransform = relativeMotion.toTF();
	std::vector<Pose> newPoses(particles.size());
	for (size_t i = 0; i < particles.size(); i++) {

	    Pose randomMotion;
	    double yaw, pitch, roll;
	    relativeMotion.getYPR(yaw, pitch, roll);

	    double& dx = relativeMotion.position.x;
	    double& dy = relativeMotion.position.y;
	    double dxy = sqrt(dx*dx + dy*dy);

	    randomMotion.position.x = ProbabilityTable.NormRand48(dx,
	            motionModel.range * fabs(dx) + motionModel.rangeXY * fabs(dxy) +
	            motionModel.slip * fabs(yaw));
	    randomMotion.position.y = ProbabilityTable.NormRand48(dy,
	            motionModel.shift * fabs(dy) + motionModel.shiftXY * fabs(dxy) +
	            motionModel.slip * fabs(yaw));

	    yaw = ProbabilityTable.NormRand48(yaw,
	            motionModel.theta * fabs(yaw) + motionModel.drift * fabs(dxy));

	    // Apply smearing.
	    if (calculateWeight) {
	        randomMotion.position.x += ProbabilityTable.NormRand48(0, motionModel.smearX);
	        randomMotion.position.y += ProbabilityTable.NormRand48(0, motionModel.smearY);
	        yaw += ProbabilityTable.NormRand48(0, motionModel.smearTheta);
	    }

	    randomMotion.setYPR(yaw, pitch, roll);
	    particles[i]->applyMotion(randomMotion.toTF(), cloud, calculateWeight, parameters.gain);
	    newPoses[i] = particles[i]->pose;
	}

	{
	    Lock lock(particlesLock);

	    motion->applyMotion(motionTransform, cloud, calculateWeight);
		particlePoses = newPoses;
		meanPose = mean->pose;
	}
}

void FastSLAMMap::updateParticles(MapCloudPtr cloud) {
	// Update motion particle
    motion->update(cloud);

	for (size_t i = 0; i < particles.size(); i++) {
	    particles[i]->update(cloud);
	}
}

void FastSLAMMap::resample() {
    // TODO: combine this with resampleAndUpdate (as these two seem to be similar methods)
	if (particles.size() != 0) {

        ParticlePtr newMean = particles[0];
        double weightMax = newMean->weight;
        double weightSum = 0;
        for (size_t i = 0; i < particles.size(); i++) {
            ParticlePtr p = particles[i];
            if (p->weight > weightMax) {
                newMean = p;
                weightMax = p->weight;
            }
            weightSum += p->weight;
        }

        std::vector<ParticlePtr> newParticles;
        std::vector<Pose> newPoses;
        std::vector<int> marks;
        marks.resize(particles.size());
        memset(&marks[0], 0, marks.size()*sizeof(int));

        newParticles.resize(parameters.numberOfParticles);
        newPoses.resize(newParticles.size());
        size_t i = 0;
        if (parameters.addMotionCopy) {
            newParticles[i] = new Particle(motion);
            ++i;
        }

        for (; i < newParticles.size(); i++) {
            double rand = ProbabilityTable.randxy(0, weightSum);
            ParticlePtr p = particles[0];
            double wCurrent = p->weight;
            size_t j = 0;
            for (++j; j < particles.size() && wCurrent <= rand; j++) {
                p = particles[j];
                wCurrent += p->weight;
            }

            j--;
            if (marks[j]) {
                newParticles[i] = new Particle(p);
                newPoses[i] = p->pose;
            } else {
                newParticles[i] = p;
                newPoses[i] = p->pose;
                marks[j] = true;
            }
        }

        {
            Lock lock(particlesLock, true);

            mean = newMean;
            particles.clear();
            particles = newParticles;
            newParticles.clear();
            particlePoses = newPoses;
            meanPose = mean->pose;
        }
	}
}

void FastSLAMMap::resampleAndUpdate(MapCloudPtr cloud) {
	if (particles.size() == 0) {
		return;
	}

	double weightMax, weightSum = 0;
	ParticlePtr newMean = particles[0];
	weightMax = newMean->weight;
	for (size_t i = 0; i < particles.size(); i++) {
		ParticlePtr p = particles[i];
		if (p->weight > weightMax) {
			newMean = p;
			weightMax = p->weight;
		}
		weightSum += p->weight;
	}

	std::vector<int> marks;
	marks.resize(particles.size());
	memset(&marks[0], 0, marks.size()*sizeof(int));
	std::vector<ParticlePtr> newParticles;
	std::vector<Pose> newPoses;

	// Update and add motion particle
	motion->update(cloud);
	size_t i = 0;
	if (parameters.addMotionCopy) {
	    ParticlePtr newParticle = new Particle(motion);
	    newParticles.push_back(newParticle);
	    newPoses.push_back(newParticle->pose);
		++i;
	}

	// Update or add new particles
	for (; i < parameters.numberOfParticles; ++i) {
		double rand = ProbabilityTable.randxy(0, weightSum);
		ParticlePtr p = particles[0];
		double wCurrent = p->weight;
		size_t j = 0;
		for (++j; j < particles.size() && wCurrent <= rand; ++j) {
			p = particles[j];
			wCurrent += p->weight;
		}

		ParticlePtr updatedParticle;
		j--;
		if (!marks[j]) {
		    updatedParticle = particles[j];
		    updatedParticle->update(cloud);
			marks[j] = true;
		} else {
		    updatedParticle = new Particle(particles[j]);
		}
        newParticles.push_back(updatedParticle);
        newPoses.push_back(updatedParticle->pose);
	}

	//ROS_INFO("FastSLAM %Zu particle updates, %Zu particle copies.\n", updates.size(), copies.size());

	// Swap in the new mean and new particles
	{
	    Lock lock(particlesLock, true);

	    mean = newMean;
		particles = newParticles;
		particlePoses = newPoses;
		meanPose = mean->pose;
	}
}

} // namespace fastslam

} // namespace crosbot
