/*
 * module.cpp
 *
 *  Created on: 20/02/2013
 *      Author: mmcgill
 */

#include <ros/ros.h>

#include <crosbot/config.hpp>
#include <crosbot_fastslam/module.hpp>
#include <geometry_msgs/PoseArray.h>

#define LOG_START   "FastSLAMModule :: "

namespace crosbot {

namespace fastslam {

/**
 * Add FastSLAM to the set of map factories that can be generated by the Crosbot GUI
 */
class FastSLAMFactory {
public:
	static MapPtr factory(ConfigElementPtr config) {
		std::string type = config->getParam(PARAM_TYPE, config->getChildName());
		std::string name = config->getParam(PARAM_NAME, "");

		if (strcasecmp(type.c_str(), "fastslam") == 0) {
			ros::NodeHandle nh;
			FastSLAMModulePtr rval = new FastSLAMModule(name);
			rval->initialize(nh);
			rval->configure(config);
			return rval;
		}
		return NULL;
	}

	FastSLAMFactory() {
		MapFactory::addFactory(&factory);
	}
};
FastSLAMFactory fastSLAMFactory;
} // namespace fastslam

} // namespace crosbot


namespace crosbot {

namespace fastslam {

FastSLAMModule::FastSLAMModule(std::string name) :
    FastSLAMMap(),
    mapFrame(DEFAULT_MAPFRAME),
    odomFrame(DEFAULT_ODOMFRAME),
    baseFrame(DEFAULT_BASEFRAME)
{
    addListener(this);
}

FastSLAMModule::~FastSLAMModule() {
    removeListener(this);
}

void FastSLAMModule::initialize(ros::NodeHandle& nh) {
    ros::NodeHandle paramNH("~");
    paramNH.param<std::string>("base_frame", baseFrame, DEFAULT_BASEFRAME);
    paramNH.param<std::string>("map_frame", mapFrame, DEFAULT_MAPFRAME);
    paramNH.param<std::string>("odom_frame", odomFrame, DEFAULT_ODOMFRAME);
    paramNH.param<std::string>("pub_history", pub_historyName, "/crosbot_fastslam/history");
    paramNH.param<std::string>("pub_map", pub_mapName, "/crosbot_fastslam/map");
    paramNH.param<std::string>("pub_particles", pub_particlesName, "/crosbot_fastslam/particles");
    paramNH.param<std::string>("sub_scan", sub_scanName, "/scan");
    paramNH.param<std::string>("sub_snap", sub_snapName, "/crosbot_fastslam/snaps");
    ROS_INFO("%s map: %s, odom: %s, base: %s", LOG_START, mapFrame.c_str(), odomFrame.c_str(), baseFrame.c_str());

    scanSub = nh.subscribe(sub_scanName, 1, &FastSLAMModule::callbackScan, this);
    snapSub = nh.subscribe(sub_snapName, 1000, &FastSLAMModule::callbackSnap, this);
    gridPub = nh.advertise<nav_msgs::OccupancyGrid>(pub_mapName, 1, true);
    historyPub = nh.advertise<nav_msgs::Path>(pub_historyName, 1, true);
    particlesPub = nh.advertise<geometry_msgs::PoseArray>(pub_particlesName, 1, true);

    // read configuration/parameters
    ConfigElementPtr config = new ROSConfigElement(paramNH);
    configure(config);

    start();
}

void FastSLAMModule::shutdown() {
    stop();

    scanSub.shutdown();
    snapSub.shutdown();
    gridPub.shutdown();
    historyPub.shutdown();
}

void FastSLAMModule::callbackScan(const sensor_msgs::LaserScanConstPtr& latestScan) {
    Pose robotPose, sensorPose;

    tf::StampedTransform transform;
    try {
        tfListener.waitForTransform(odomFrame,
                latestScan->header.frame_id, latestScan->header.stamp,
                ros::Duration(DEFAULT_MAXWAIT4TRANSFORM));

        // Load transform laser to base
        tfListener.lookupTransform(baseFrame,
                latestScan->header.frame_id, latestScan->header.stamp, transform);
        sensorPose = transform;

        // Load transform base to odom estimate
        tfListener.lookupTransform(odomFrame, baseFrame,
                latestScan->header.stamp, transform);
        robotPose = transform;
    } catch (tf::TransformException& ex) {
        ROS_WARN("%s Problem obtaining transform (%s)\n", LOG_START, ex.what());
        return;
    }

    // Load new laser scan point cloud
    PointCloud pc(latestScan);
    newPointCloud(new PointCloud(baseFrame, pc, sensorPose.toTF()), robotPose, sensorPose);
}

void FastSLAMModule::callbackSnap(const crosbot_map::SnapMsgConstPtr& snap) {
    // TODO: convert all point clouds to base frame
    addTag(new Snap(snap));
}

void FastSLAMModule::publishTransform(const Pose& robot, const Pose& odometry, const Time& time) {
    geometry_msgs::TransformStamped ts;
    ts.header.stamp = time.toROS();
    ts.header.frame_id = mapFrame;
    ts.child_frame_id = odomFrame;

    Pose pose = robot.toTF() * odometry.toTF().inverse();

    ts.transform.translation.x = pose.position.x;
    ts.transform.translation.y = pose.position.y;
    ts.transform.translation.z = pose.position.z;
    ts.transform.rotation = pose.orientation.toROS();
    tfBroadcast.sendTransform(ts);
}

void FastSLAMModule::motionTracked(MapPtr map) {
    // TODO: Update not just pose, but everything
    ParticlePtr newMean = new Particle(mean, true);
    MapCloudPtr update = newMean->getLatestUpdate();
    if (update != NULL) {
        Pose pose = newMean->getPose();
        publishTransform(newMean->getPose(), update->odometry, update->timestamp);
    }
}

void FastSLAMModule::mapUpdated(MapPtr map) {
    // TODO: Periodically call/republish map and particles - not just on new scan data/map update
    ParticlePtr newMean = new Particle(mean);

    // Use common "now" timestamp
    ros::Time tNow = ros::Time::now();

    // Publish transform
    MapCloudPtr update = newMean->getLatestUpdate();
    if (update != NULL) {
        ROS_INFO("%s map updating", LOG_START);
        Pose pose = newMean->getPose();
        publishTransform(newMean->getPose(), update->odometry, update->timestamp);
    }

    // Publish Map Occupancy Grid
    if (gridPub.getNumSubscribers() > 0) {
        nav_msgs::OccupancyGridPtr grid = asOccupancyGrid(newMean);
        if (grid != NULL) {
            grid->header.frame_id = mapFrame;
            grid->header.stamp = tNow;
            gridPub.publish(grid);
        }
    }

    // Publish history of robot path
    if (historyPub.getNumSubscribers() > 0) {
        nav_msgs::PathPtr rosPath(new nav_msgs::Path());
        rosPath->header.frame_id = mapFrame;
        rosPath->header.stamp = newMean->getLatestUpdate()->timestamp.toROS();

        PathPtr path = newMean->getPath();
        rosPath->poses.resize(path->path.size());
        for (size_t i = 0; i < path->path.size(); ++i) {
            geometry_msgs::PoseStamped& pose = rosPath->poses[i];
            pose.pose = path->path[i].toROS();
            pose.header.frame_id = mapFrame;
            pose.header.stamp = (i < path->timestamps.size())?(path->timestamps[i].toROS()):ros::Time(0);
        }

        historyPub.publish(rosPath);
    }

    // Publish particles
    if (particlesPub.getNumSubscribers() > 0) {
        geometry_msgs::PoseArray particlesMsg;
        particlesMsg.header.stamp = tNow;
        particlesMsg.header.frame_id = mapFrame;
        for (Pose& pose : particlePoses) {
            particlesMsg.poses.push_back(pose.toROS());
        }
        particlesPub.publish(particlesMsg);
    }
}

void FastSLAMModule::tagAdded(MapPtr map, TagPtr tag) {
    // TODO: FastSLAMModule::tagAdded()
}

void FastSLAMModule::tagChanged(MapPtr map, TagPtr tag) {
    // TODO: FastSLAMModule::tagChanged()
}

} // namespace fastslam

} // namespace crosbot
