/*
 * occGridMerger_node.cpp
 *
 *  Created on: 07/03/2017
 *      Author: timothyw
 */


#include <ros/ros.h>

#include <crosbot/config.hpp>
#include <crosbot/thread.hpp>
#include <crosbot_map/occGridMerger.hpp>
#include <ros/publisher.h>
#include <ros/subscriber.h>

#include <string>

#define LOG_START "OccGridMergerNode ::"

namespace crosbot_map {

class OccGridMergerNode : public OccGridMerger {
private:
    // Configuration
    std::string mapSub_name;
    std::string laserSub_name;
    std::string combinedPub_name;
    double publishRate;

    // Publishers/Subscribers
    ros::Publisher  pub_combined;
    ros::Subscriber sub_map;
    ros::Subscriber sub_laser;
    ros::Publisher  pub_cloudDebug;

    bool operating;

    // Update thread
    class CombineThread : public crosbot::Thread {
    public:
        OccGridMergerNode& node;
        CombineThread(OccGridMergerNode& node) : node(node) {}

        void run() {
            node.combineThread();
        }
    };
    CombineThread thread;

public:
    OccGridMergerNode() :
        OccGridMerger(),
        operating(false),
        thread(*this),
        publishRate(1)
    {};
    virtual ~OccGridMergerNode() {};

    void callback_map(const nav_msgs::OccupancyGridConstPtr& map) {
        updateMap(map);
    };

    void callback_laser(const sensor_msgs::LaserScanConstPtr scan) {
        updateLaser(scan);
    };

    void configure() {
        ros::NodeHandle nh("~");
        crosbot::ConfigElementPtr cfg = new crosbot::ROSConfigElement(nh);

        combinedPub_name = cfg->getParam("combined_pub", "combined");
        mapSub_name = cfg->getParam("map_sub", "/map");
        laserOccupiedProbability = cfg->getParamAsDouble("laser_occupied_prob", 100);
        laserMinDist = cfg->getParamAsDouble("laser_min_dist", 0.1);
        laserSub_name = cfg->getParam("laser_sub", "/scan");
        publishRate = cfg->getParamAsDouble("publish_rate", 10);

        clear();
    };

    void start() {
        // Subscribe/Advertise
        ros::NodeHandle nh;
        pub_combined = nh.advertise<nav_msgs::OccupancyGrid>(combinedPub_name, 1);
        sub_map = nh.subscribe(mapSub_name, 1, &OccGridMergerNode::callback_map, this);
        sub_laser = nh.subscribe(laserSub_name, 1, &OccGridMergerNode::callback_laser, this);
        pub_cloudDebug = nh.advertise<sensor_msgs::PointCloud>("merger_cloud_debug", 1);

        clear();
        operating = true;
        thread.start();
    };

    void shutdown() {
        operating = false;

        pub_combined.shutdown();
        sub_map.shutdown();
        sub_laser.shutdown();
        pub_cloudDebug.shutdown();

        clear();
    };

    void combineThread() {
        ros::Rate rate(publishRate);
        while (operating) {
            // only process if there are subscribers
            if (pub_combined.getNumSubscribers() > 0) {
                // Get next map
                //ROS_INFO("%s get map", LOG_START);
                nav_msgs::OccupancyGridConstPtr map = generateMap();

                // Publish
                if (map != NULL) {
                    //ROS_INFO("%s publish map", LOG_START);
                    pub_combined.publish(map);
                } else {
                    //ROS_INFO("%s map is NULL", LOG_START);
                }

                if (debugCloud != NULL && pub_cloudDebug.getNumSubscribers() > 0) {
                    pub_cloudDebug.publish(debugCloud->toROS1());
                }
            }

            rate.sleep();
        }
    }
};

} // namespace crosbot_map

int main(int argc, char** argv) {
    ros::init(argc, argv, "crosbot_map_occupancy_grid_merger");

    crosbot_map::OccGridMergerNode node;
    node.configure();
    node.start();

    while (ros::ok()) {
        ros::spin();
    }
    node.shutdown();

    return 0;
}

