/*
 * occGridMerger.cpp
 *
 *  Created on: 07/03/2017
 *      Author: timothyw
 */


#include <ros/ros.h>

#include <crosbot/geometry/pointCloud.hpp>
#include <crosbot_map/occGridMerger.hpp>

#define LOG_START                   "OccGridMerger ::"

namespace crosbot_map {

OccGridMerger::OccGridMerger() :
    crosbot::HandledObject(),
    map(NULL),
    scan(NULL),
    laserOccupiedProbability(100),
    laserMinDist(0.1)
{}

OccGridMerger::~OccGridMerger() {
};

void OccGridMerger::clear() {
    crosbot::Lock dataLock(dataMutex);
    map = NULL;
    scan = NULL;
}


void OccGridMerger::updateMap(const nav_msgs::OccupancyGridConstPtr& map) {
    crosbot::Lock dataLock(dataMutex);
    this->map = map;
}

void OccGridMerger::updateLaser(const sensor_msgs::LaserScanConstPtr scan) {
    crosbot::Lock dataLock(dataMutex);
    this->scan = scan;
}

nav_msgs::OccupancyGridConstPtr OccGridMerger::generateMap() {
    if (map == NULL || scan == NULL) {
        return nav_msgs::OccupancyGridConstPtr(NULL);
    }

    // Load copy of maps
    nav_msgs::OccupancyGridConstPtr map;
    sensor_msgs::LaserScanConstPtr scan;
    {
        crosbot::Lock mapsLock(dataMutex);
        map = this->map;
        scan = this->scan;
    }

    // Generate base Occupancy Grid
    nav_msgs::OccupancyGrid* combined(new nav_msgs::OccupancyGrid());
    combined->header.frame_id = map->header.frame_id;
    combined->info.height = map->info.height;
    combined->info.width = map->info.width;
    combined->info.resolution = map->info.resolution;
    combined->info.origin = map->info.origin;
    combined->header.stamp = (map->header.stamp > scan->header.stamp) ? map->header.stamp : scan->header.stamp;
    combined->info.map_load_time = combined->header.stamp;

    // Data
    //combined->data.resize(map->data.size());
    combined->data.insert(combined->data.end(), map->data.begin(), map->data.end());

    // Laser scan transform
    tf::StampedTransform transform;
    try {
        tfListener.waitForTransform(map->header.frame_id, scan->header.frame_id, scan->header.stamp,
                ros::Duration(0.2));
        tfListener.lookupTransform(map->header.frame_id, scan->header.frame_id, scan->header.stamp, transform);
    } catch (tf::TransformException& ex) {
        //ROS_WARN("%s Error getting transform. (%s)\n", LOG_START, ex.what());
        nav_msgs::OccupancyGridConstPtr rval(combined);
        return rval;
    }

    // Remove laser points which are too close
    sensor_msgs::LaserScanPtr scanAlt(new sensor_msgs::LaserScan());
    scanAlt->header = scan->header;
    scanAlt->angle_min = scan->angle_min;
    scanAlt->angle_max = scan->angle_max;
    scanAlt->angle_increment = scan->angle_increment;
    scanAlt->time_increment = scan->time_increment;
    scanAlt->scan_time = scan->scan_time;
    scanAlt->range_min = laserMinDist;
    scanAlt->range_max = 5.0;
    scanAlt->ranges.insert(scanAlt->ranges.end(), scan->ranges.begin(), scan->ranges.end());

    crosbot::PointCloudPtr scanPc = new crosbot::PointCloud(scanAlt);
    //ROS_INFO("min: %.2lf", scanAlt->range_min);

    // Put laser scan in frame of reference

    crosbot::PointCloudPtr laserCloud = new crosbot::PointCloud(map->header.frame_id, *scanPc, transform);
    debugCloud = scanPc;
    //debugCloud = laserCloud;

    // Add laser data in
    uint32_t x;
    uint32_t y;
    for (size_t i = 0; i < laserCloud->cloud.size(); ++i) {
        const crosbot::Point& p = laserCloud->cloud[i];

        x = (p.x - combined->info.origin.position.x) / combined->info.resolution;
        y = (p.y - combined->info.origin.position.y) / combined->info.resolution;

        if (x >= 0 && x < combined->info.width && y >= 0 && y < combined->info.height) {
            uint32_t index = y * combined->info.width + x;
            combined->data[index] = laserOccupiedProbability;
        }
    }

    nav_msgs::OccupancyGridConstPtr rval(combined);
    return rval;
}

} // namespace crosbot_map
