/*
 * geotiff.cpp
 *
 *  Created on: 08/07/2016
 *      Author: rescue
 */


#include <ros/ros.h>

#include <crosbot_map_geotiff/geotiff.hpp>

#include <crosbot/data/time.hpp>
#include <crosbot/geometry/poses.hpp>
#include <sensor_msgs/Image.h>
#include <sensor_msgs/image_encodings.h>
#include <QColor>
#include <QImage>
#include <QPainter>
#include <vector>

#define LOG_START           "CrosbotGeotiff ::"

namespace crosbot_map_geotiff {

CrosbotGeotiff::CrosbotGeotiff() :
    operating(false),
    frequency(0),
    haveMap(false),
    history(NULL),
    map(NULL)

{
}

CrosbotGeotiff::~CrosbotGeotiff() {
}

void CrosbotGeotiff::configure() {
    ros::NodeHandle paramNh("~");

    frequency = paramNh.param<double>("frequency", 1.0);
    markers_topic = paramNh.param<std::string>("markers_topic", "/markers");
    map_topic = paramNh.param<std::string>("map_topic", "/map");
    mapImg_topic = paramNh.param<std::string>("mapImg_topic", "/map/geotiff");
    path_topic = paramNh.param<std::string>("path_topic", "/path");
    tiff_filename = paramNh.param<std::string>("tiff_filename", "/home/rescue/geotiff.tiff");
}

void CrosbotGeotiff::start() {
    // Set operating
    operating = true;

    // Subscribe to topics
    ros::NodeHandle nh("~");
    mapImgPub = nh.advertise<sensor_msgs::Image>(mapImg_topic, 1);
    markersSub = nh.subscribe(markers_topic, 1, &CrosbotGeotiff::callback_markers, this);
    mapSub = nh.subscribe(map_topic, 1, &CrosbotGeotiff::callback_map, this);
    pathSub = nh.subscribe(path_topic, 1, &CrosbotGeotiff::callback_path, this);

    history = NULL;
    map = NULL;

    Thread::start();
}

void CrosbotGeotiff::stop() {
    crosbot::Lock lock(processMutex);

    // Unsubscribe
    mapImgPub.shutdown();
    mapSub.shutdown();

    // Terminate
    operating = false;

    // Cleanup
    haveMap = false;
    map = NULL;
    history = NULL;
}

void CrosbotGeotiff::run() {
    while (operating) {

        if (haveMap) {
            crosbot::Lock lock(processMutex);
            //ROS_INFO("%s producing geotiff", LOG_START);

            // Generate Qt geotiff object
            QImage *qimg = produceQtGeotiff();

            // Publish geotiff as ros image
            QImage rgbImg = qimg->convertToFormat(QImage::Format_RGB888);
            const uchar *rgbData = rgbImg.constBits();
            int rgbLength = rgbImg.height() * rgbImg.width() * 3;

            sensor_msgs::Image rosImg;
            rosImg.header.frame_id = map->header.frame_id;
            rosImg.header.stamp = map->header.stamp;
            rosImg.encoding = sensor_msgs::image_encodings::RGB8;
            rosImg.step = 3;
            rosImg.height = qimg->height();
            rosImg.width = qimg->width();
            rosImg.data.assign(rgbData, rgbData + rgbLength);
            mapImgPub.publish(rosImg);

            // Export tiff image
            if (!qimg->save(QString(tiff_filename.c_str()), "TIFF")) {
                char eMsg[tiff_filename.size() + 128];

                sprintf(eMsg, "Could not save raw TIFF to file %s.", tiff_filename.c_str());
                ROS_ERROR("geo_logger: Problem saving map to TIFF. (file %s)\n", tiff_filename.c_str());
            }

            // Cleanup
            delete qimg;
        }

        sleep(frequency);
    }
}

void CrosbotGeotiff::callback_markers(const visualization_msgs::MarkerConstPtr marker) {
    // Update marker
    crosbot::Lock lock(processMutex);
    //ROS_INFO("%s marker: %d", LOG_START, marker->id);

    if (marker->action == visualization_msgs::Marker::ADD || marker->action == visualization_msgs::Marker::MODIFY) {
        //ROS_INFO("%s adding marker: %d", LOG_START, marker->id);
        markers[marker->id] = *marker;
    } else if (marker->action == visualization_msgs::Marker::DELETE) {
        if (markers.find(marker->id) != markers.end()) {
            markers.erase(marker->id);
        }
    }
}


void CrosbotGeotiff::callback_map(const nav_msgs::OccupancyGridConstPtr map) {
    //ROS_INFO("%s got new map", LOG_START);
    crosbot::Lock lock(processMutex);
    this->map = map;
    haveMap = true;
}

void CrosbotGeotiff::callback_path(const nav_msgs::PathConstPtr path) {
    crosbot::Lock lock(processMutex);
    history = path;
}

// TODO: finish all required elements
QImage* CrosbotGeotiff::produceQtGeotiff() {
    if (map == NULL || map->info.width == 0 || map->info.height == 0) {
        ROS_ERROR("GeoTiff: Cannot draw an empty map.\n");
        return NULL;
    }

    // create image
    unsigned int imageWidth = map->info.width;
    unsigned int imageHeight = map->info.height;

    QImage geotiffImage(imageWidth, imageHeight, QImage::Format_RGB32);
    QPainter painter(&geotiffImage);

    // paint unexplored checker board
    int checkerBoardPixelWidth = (int) round(GeoTIFFConstants::unexploredCheckerBoardSize / map->info.resolution);
    int checkerBoardPixelHeight = (int) round(GeoTIFFConstants::unexploredCheckerBoardSize / map->info.resolution);

    for (unsigned int x = 0; x < imageWidth; x += checkerBoardPixelWidth) {
        for (unsigned int y = 0; y < imageHeight; y += checkerBoardPixelHeight) {
            Colour4f c;
            if (((x / checkerBoardPixelWidth + y / checkerBoardPixelHeight) & 0x01) == 0) {
                c = GeoTIFFConstants::unexploredLight;
            } else {
                c = GeoTIFFConstants::unexploredDark;
            }
            painter.fillRect(QRect(x, y, checkerBoardPixelWidth, checkerBoardPixelHeight),
                    QColor((int)c.r, (int)c.g, (int)c.b, 255));
        }
    }

    // paint mapped areas
    int gridWidth = (int)round(GeoTIFFConstants::exploredGridSize / map->info.resolution);
    int gridHeight = (int)round(GeoTIFFConstants::exploredGridSize / map->info.resolution);

    QColor occupied((int)GeoTIFFConstants::wallAndObstacleColour.r,
            (int)GeoTIFFConstants::wallAndObstacleColour.g,
            (int)GeoTIFFConstants::wallAndObstacleColour.b);
    QColor grid((int)GeoTIFFConstants::exploredGrid.r,
            (int)GeoTIFFConstants::exploredGrid.g,
            (int)GeoTIFFConstants::exploredGrid.b);

    QColor clearedMin((int)GeoTIFFConstants::clearedAreaLowConfidence.r,
            (int)GeoTIFFConstants::clearedAreaLowConfidence.g,
            (int)GeoTIFFConstants::clearedAreaLowConfidence.b);
    QColor clearedMax((int)GeoTIFFConstants::clearedAreaHighConfidence.r,
            (int)GeoTIFFConstants::clearedAreaHighConfidence.g,
            (int)GeoTIFFConstants::clearedAreaHighConfidence.b);
    Colour4f clearedDiff(GeoTIFFConstants::clearedAreaHighConfidence.r - GeoTIFFConstants::clearedAreaLowConfidence.r,
            GeoTIFFConstants::clearedAreaHighConfidence.g - GeoTIFFConstants::clearedAreaLowConfidence.g,
            GeoTIFFConstants::clearedAreaHighConfidence.b - GeoTIFFConstants::clearedAreaLowConfidence.b);

    QColor freeMin((int)GeoTIFFConstants::searchedAreaLowConfidence.r,
            (int)GeoTIFFConstants::searchedAreaLowConfidence.g,
            (int)GeoTIFFConstants::searchedAreaLowConfidence.b);
    QColor freeMax((int)GeoTIFFConstants::searchedAreaHighConfidence.r,
            (int)GeoTIFFConstants::searchedAreaHighConfidence.g,
            (int)GeoTIFFConstants::searchedAreaHighConfidence.b);
    Colour4f freeDiff(GeoTIFFConstants::searchedAreaHighConfidence.r - GeoTIFFConstants::searchedAreaLowConfidence.r,
            GeoTIFFConstants::searchedAreaHighConfidence.g - GeoTIFFConstants::searchedAreaLowConfidence.g,
            GeoTIFFConstants::searchedAreaHighConfidence.b - GeoTIFFConstants::searchedAreaLowConfidence.b);

    for (uint32_t y = 0; y < imageHeight; ++y) {
        for (uint32_t x = 0; x < imageWidth; ++x) {
            QRect cell(x, y, 1, 1);

            const int8_t& v = map->data[y * imageWidth + x];
            if (v > 50) {
                painter.fillRect(cell, occupied);
            } else if (v != -1) {
                // TODO: find out what crosbot_graphslam unobserved value for v iss
                float conf = v / (float)100.0;
                int r = freeMin.red() + (int)(conf * freeDiff.r),
                        g = freeMin.green() + (int)(conf * freeDiff.g),
                        b = freeMin.blue() + (int)(conf * freeDiff.b);
                QColor c(r, g, b);
                painter.fillRect(cell, c);
            }
        }
    }

    float minVisX = map->info.origin.position.x;
    float minVisY = map->info.origin.position.y;
    //ROS_INFO("minXY (%.3f, %.3f) at resolution %.2lf\n", minVisX, minVisY, map->info.resolution);

    // paint robot path
    if (history != NULL && history->poses.size() > 0) {
        float penSize = GeoTIFFConstants::robotPathThickness / map->info.resolution;
        if (penSize < 1)
            penSize = 1;
        QPen penPath(QColor((int)GeoTIFFConstants::robotPathColour.r,
                (int)GeoTIFFConstants::robotPathColour.g,
                        (int)GeoTIFFConstants::robotPathColour.b));
        penPath.setWidthF(penSize);
        painter.setPen(penPath);
        crosbot::Pose startPose = history->poses[0];
        // get pixel for xy coord.
        crosbot::Pose p = startPose;

        QPoint start((p.position.x - minVisX) / map->info.resolution, (p.position.y - minVisY) / map->info.resolution);

        for (unsigned int i = 1; i < history->poses.size(); i++) {
            p = history->poses[i];
            QPoint end((p.position.x - minVisX) / map->info.resolution, (p.position.y - minVisY) / map->info.resolution);
            painter.drawLine(start, end);
            start = end;
        }

        QPen penRobot(QColor((int)GeoTIFFConstants::initialPositionColour.r,
                (int)GeoTIFFConstants::initialPositionColour.g,
                (int)GeoTIFFConstants::initialPositionColour.b));
        penRobot.brush().setStyle(Qt::SolidPattern);
        painter.setPen(penRobot);
        painter.setBrush(penRobot.brush());
        float robotSizePixels = GeoTIFFConstants::robotPositionSize/map->info.resolution;

        {{
                // paint initial robot positions
                crosbot::Pose p = startPose;
                double yaw, pitch, roll;
                p.getYPR(yaw, pitch, roll);

                float cX = (p.position.x - minVisX) / map->info.resolution,
                        cY = (p.position.y - minVisY) / map->info.resolution;
                float cosYawRobotSize = cos(yaw)*robotSizePixels,
                        sinYawRobotSize = sin(yaw)*robotSizePixels;

                QPointF polygon[4];
                // front
                polygon[0] = QPointF(cX + cosYawRobotSize, cY + sinYawRobotSize);
                // right
                polygon[1] = QPointF(cX - (sinYawRobotSize + cosYawRobotSize),
                        cY - (sinYawRobotSize + cosYawRobotSize));
                // back
                polygon[2] = QPointF(cX - cosYawRobotSize/2, cY - sinYawRobotSize/2);
                // left
                polygon[3] = QPointF(cX - (sinYawRobotSize + cosYawRobotSize),
                        cY + (sinYawRobotSize + cosYawRobotSize));

                painter.drawPolygon(polygon, 4);
        }}
    }

    // The painted image is upside down so mirror it
    QImage* rval = new QImage();
//  QImage correctGeotiffImage = geotiffImage.mirrored(false, true);
//  QPainter textPainter(&correctGeotiffImage);
    *rval = geotiffImage.mirrored(false, true);

    QPainter textPainter(rval);
    //ROS_INFO("%s active: %d", LOG_START, textPainter.isActive());


    // paint victim/hazard labels
    QFontMetrics fm = textPainter.fontMetrics();
    int minLabelSize = fm.height() - fm.descent();


    QPen penVictim(QColor((int)GeoTIFFConstants::victimColour.r,
            (int)GeoTIFFConstants::victimColour.g,
            (int)GeoTIFFConstants::victimColour.b));
    penVictim.brush().setStyle(Qt::SolidPattern);
    QPen penVictimUnconfirmed(QColor((int)GeoTIFFConstants::victimUnconfirmedColour.r,
            (int)GeoTIFFConstants::victimUnconfirmedColour.g,
            (int)GeoTIFFConstants::victimUnconfirmedColour.b));
    penVictimUnconfirmed.brush().setStyle(Qt::SolidPattern);

    float victimRadius = GeoTIFFConstants::victimMarkerDiameter / map->info.resolution /2;
    if (victimRadius*2 < minLabelSize) {
        victimRadius = (minLabelSize + 1) / 2;
    }
    float hazardXYSize = sqrt(GeoTIFFConstants::hazardMarkerSide*GeoTIFFConstants::hazardMarkerSide/2)/map->info.resolution;
    if (hazardXYSize*2 < minLabelSize) {
        hazardXYSize = (minLabelSize + 1) / 2;
    }

    // TODO: incorporate snaps again - using only visualisation markers at present
    int victimCount = 0, hazardCount = 0;
    std::string victimDetails, hazardDetails;
    for (std::map<int, visualization_msgs::Marker>::iterator it = markers.begin(); it != markers.end(); ++it) {
        visualization_msgs::Marker& marker = it->second;

        // TODO: at present this assumes markers are in map frame
        crosbot::Pose targetPose = marker.pose;

        float cX = (targetPose.position.x - minVisX) / map->info.resolution;
        float cY = (targetPose.position.y - minVisY) / map->info.resolution;
        cY = geotiffImage.height() - cY;

        //ROS_INFO("Target (%.3lf, %.3lf) at pixel (%.0lf, %.0lf)\n", targetPose.position.x, targetPose.position.y, cX, cY);
        //ROS_INFO("minXY (%.3f, %.3f) at resolution %.2lf\n", minVisX, minVisY, map->info.resolution);

        char iStr[64];
        std::string markText;
        sprintf(iStr, "%d", marker.id);
        markText = iStr;

        QPen penHazard(QColor(marker.color.r, marker.color.g, marker.color.b));
        penHazard.brush().setStyle(Qt::SolidPattern);
        textPainter.setPen(penHazard);
        textPainter.setBrush(penHazard.brush());
        QPointF diamond[4];
        diamond[0] = QPointF(cX, cY + hazardXYSize);
        diamond[1] = QPointF(cX + hazardXYSize, cY);
        diamond[2] = QPointF(cX, cY - hazardXYSize);
        diamond[3] = QPointF(cX - hazardXYSize, cY);

        crosbot::Time timestamp = marker.header.stamp;
        hazardDetails += std::string(iStr) + ";" + timestamp.format("%H:%M:%S;") + marker.text + ";";
        sprintf(iStr, "%.3lf;", targetPose.position.x);
        hazardDetails += std::string(iStr);
        sprintf(iStr, "%.3lf;", targetPose.position.y);
        hazardDetails += std::string(iStr);
        sprintf(iStr, "%.3lf\n", targetPose.position.z);
        hazardDetails += std::string(iStr);

        textPainter.drawPolygon(diamond, 4);

        textPainter.setPen(QColor(255,255,255));
        float dX = fm.width(QString(markText.c_str())) / 2.0;
        float dY = fm.height() / 2.0 - fm.descent();

        textPainter.drawText(QPointF(cX - dX, cY + dY), QString(markText.c_str()));
    }
    /*for (unsigned int i = 0; i < snaps.size(); i++) {
        SnapPtr snap = new Snap(snaps[i]);
        if (snap == NULL)
            continue;
        if (snap->status == Snap::CONFIRMED || snap->status == Snap::UNCONFIRMED) {
            Pose targetPose = snap->robot.toTF() * snap->pose.toTF();

            float cX = (targetPose.position.x - minVisX) / map->info.resolution,
                    cY = (targetPose.position.y - minVisY) / map->info.resolution;
            cY = geotiffImage.height() - cY;

            LOG("Target (%.3lf, %.3lf) at pixel (%.0lf, %.0lf)\n",
                    targetPose.position.x, targetPose.position.y, cX, cY);

            LOG("minXY (%.3f, %.3f) at resolution %.2lf\n", minVisX, minVisY, map->info.resolution);

            char iStr[64];
            std::string markText;
            if (snap->type == Snap::VICTIM) {
                sprintf(iStr, "%d", ++victimCount);
                markText = iStr;
                if (snap->status == Snap::CONFIRMED) {
                    textPainter.setPen(penVictim);
                    textPainter.setBrush(penVictim.brush());
                } else {
                    textPainter.setPen(penVictimUnconfirmed);
                    textPainter.setBrush(penVictimUnconfirmed.brush());
                }

                victimDetails += std::string(iStr) + ", " + targetPose.position.toString() + ", " + snap->description + "\n";

                textPainter.drawEllipse(QPointF(cX, cY), victimRadius,victimRadius);
            } else if (snap->type == Snap::LANDMARK){
                sprintf(iStr, "%d", ++hazardCount);
                markText = iStr;

                textPainter.setPen(penHazard);
                textPainter.setBrush(penHazard.brush());
                QPointF diamond[4];
                diamond[0] = QPointF(cX, cY + hazardXYSize);
                diamond[1] = QPointF(cX + hazardXYSize, cY);
                diamond[2] = QPointF(cX, cY - hazardXYSize);
                diamond[3] = QPointF(cX - hazardXYSize, cY);

                hazardDetails += std::string(iStr) + ";" + snap->timestamp.format("%H:%M:%S;") + snap->description + ";";
                sprintf(iStr, "%.3lf;", targetPose.position.x);
                hazardDetails += std::string(iStr);
                sprintf(iStr, "%.3lf;", targetPose.position.y);
                hazardDetails += std::string(iStr);
                sprintf(iStr, "%.3lf\n", targetPose.position.z);
                hazardDetails += std::string(iStr);

                textPainter.drawPolygon(diamond, 4);
            }

            textPainter.setPen(QColor(255,255,255));
            float dX = fm.width(QString(markText.c_str())) / 2.0, dY = fm.height() / 2.0 - fm.descent();

            textPainter.drawText(QPointF(cX - dX, cY + dY), QString(markText.c_str()));
        }
    }*/
    /*if (hazardCount > 0) {
        snapDetails += "CASualty, Australia\n" + Time::now().format("%Y-%m-%d; %H:%M:%S") +
                "\nMission 4\n\n" + hazardDetails;
    }*/

    // Scale
    textPainter.setPen(QColor((int)GeoTIFFConstants::scaleColour.r,
            (int)GeoTIFFConstants::scaleColour.g,
            (int)GeoTIFFConstants::scaleColour.b));
    int endCapLen = 6;
    QPoint scaleTop(checkerBoardPixelWidth/2, checkerBoardPixelHeight);
    QPoint scaleBottom(scaleTop.x(), scaleTop.y()+checkerBoardPixelHeight);
    textPainter.drawLine(scaleTop, scaleBottom);
    textPainter.drawLine(QPoint(scaleTop.x() - endCapLen / 2,  scaleTop.y()),
            QPoint(scaleTop.x() + endCapLen / 2,  scaleTop.y()));
    textPainter.drawLine(QPoint(scaleBottom.x() - endCapLen / 2,  scaleBottom.y()),
            QPoint(scaleBottom.x() + endCapLen / 2,  scaleBottom.y()));
    QPoint scaleText(scaleTop.x() + 5,
            (scaleTop.y() + scaleBottom.y()) / 2 + (fm.height() / 2 - fm.descent()));
    textPainter.drawText(scaleText, "1m");

    // Orientation
    // MM: The robocup rescue instructions call for the x axis to go up and y axis to go left.
    //     We accomplish this by swapping the x and y values and then negating y.
    textPainter.setPen(QColor((int)GeoTIFFConstants::orientationColour.r,
            (int)GeoTIFFConstants::orientationColour.g,
            (int)GeoTIFFConstants::orientationColour.b));
    QPoint orientOrigin(checkerBoardPixelWidth*3, checkerBoardPixelHeight*2);
    QPoint xEnd(orientOrigin.x(), orientOrigin.y()-checkerBoardPixelHeight);
    QPoint yEnd(orientOrigin.x()+checkerBoardPixelWidth, orientOrigin.y());
    textPainter.drawLine(orientOrigin, xEnd);
    textPainter.drawLine(orientOrigin, yEnd);
    textPainter.drawLine(xEnd, QPoint(xEnd.x() - 2, xEnd.y() + 2));
    textPainter.drawLine(xEnd, QPoint(xEnd.x() + 2, xEnd.y() + 2));
    textPainter.drawLine(yEnd, QPoint(yEnd.x() - 2, yEnd.y() - 2));
    textPainter.drawLine(yEnd, QPoint(yEnd.x() - 2, yEnd.y() + 2));
    int dY = (fm.height() - fm.descent() -fm.ascent()) / 2;
    int dX = fm.width("y") / 2;
    QPoint yText((orientOrigin.x() + yEnd.x()) / 2 - dX - 2, orientOrigin.y() + fm.height() - fm.descent());
    QPoint xText(orientOrigin.x() - 12, (orientOrigin.y() + xEnd.y()) / 2 + dY + 2);
    textPainter.drawText(xText, "y");
    textPainter.drawText(yText, "x");

    /*if (title == "") {
        textPainter.setPen(QColor((int)GeoTIFFConstants::textColour.r,
                (int)GeoTIFFConstants::textColour.g,
                (int)GeoTIFFConstants::textColour.b));
        QFontMetrics fm = textPainter.fontMetrics();
        textPainter.drawText(checkerBoardPixelWidth, checkerBoardPixelHeight - fm.descent(), QString(title.c_str()));
    }

    char data[64*6];
    geoData = "";
    sprintf(data, "%f\n", map->info.resolution);    // x resolution (meters/pixel)
    geoData += data;
    sprintf(data, "%f\n", 0.0);         // translation
    geoData += data;
    sprintf(data, "%f\n", 0.0);         // rotation
    geoData += data;
    sprintf(data, "%f\n", map->info.resolution);    // -y resolution (meters/pixel)
    geoData += data;
    sprintf(data, "%f\n", minVisX); // upper left cell
    geoData += data;
    sprintf(data, "%f\n", minVisY + (imageHeight * map->info.resolution));  // upper left cell
    geoData += data;*/

    return rval;
}

} // namespace crosbot_map_geotiff

